#ifndef _itkekADMeanSquaresImageToImageMetric_txx
#define _itkekADMeanSquaresImageToImageMetric_txx

#include "itkekADMeanSquaresImageToImageMetric.h"

#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkLinearInterpolateImageFunction.h"

#include "itkekADRevolveTransformBase.h"
#include "itkekMacro.h"
#include <time.h>

namespace itkek
{
   
/*
 * Constructor
 */
template <class TFixedImage, class TMovingImage> 
ADMeanSquaresImageToImageMetric<TFixedImage,TMovingImage>
//ADMeanSquaresImageToImageMetric
::ADMeanSquaresImageToImageMetric()
{
   itkDebugMacro("Constructor");
}


/*
 * Get the match Measure
 */
// This is the same as the function after it but without the 
// automatic diff stuff.
template <class TFixedImage, class TMovingImage> 
typename ADMeanSquaresImageToImageMetric<TFixedImage,TMovingImage>::MeasureType
ADMeanSquaresImageToImageMetric<TFixedImage,TMovingImage>
::GetValue( const TransformParametersType & parameters ) const
{
//   EVAL( unsigned int start_time2 = time(0) );
	EVAL( unsigned int start_time = clock() );



   itkDebugMacro("GetValue( " << parameters << " ) ");

   FixedImageConstPointer fixedImage = this->GetFixedImage();
   MovingImageConstPointer movingImage = this->GetMovingImage();

   if( !fixedImage ) 
   {
      itkExceptionMacro( << "Fixed image has not been assigned" );
   }

//   typename FixedImageType::RegionType::SizeType s = 
//      fixedImage->GetLargestPossibleRegion().GetSize();
//   const unsigned int numberOfVoxelsInImage = s[0] * s[1] * s[2];

   const unsigned int numberOfVoxelsInImage = 
      fixedImage->GetLargestPossibleRegion().GetNumberOfPixels();

   typedef  itk::ImageRegionConstIteratorWithIndex<FixedImageType> FixedIteratorType;


   FixedIteratorType ti( fixedImage, this->GetFixedImageRegion() );

   typename FixedImageType::IndexType index;

   MeasureType measure = itk::NumericTraits< MeasureType >::Zero;

   this->SetTransformParameters( parameters );

   while(!ti.IsAtEnd())
   {

      index = ti.GetIndex();
    
      typename Superclass::InputPointType inputPoint;
      fixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

      typename Superclass::OutputPointType transformedPoint = 
	 this->m_Transform->TransformPoint( inputPoint );

      if( this->m_Interpolator->IsInsideBuffer( transformedPoint ) )
      {
	 const RealType movingValue = this->m_Interpolator->Evaluate( transformedPoint);
	 const RealType fixedValue   = ti.Get();
	 const RealType diff = movingValue - fixedValue; 
	 measure += diff * diff; 

      }
      else
      {
       //const RealType movingValue  = 0.0;
       const RealType fixedValue   = ti.Get();
       //const RealType diff = movingValue - fixedValue; 
       //measure += diff * diff; 
       measure += fixedValue * fixedValue;
      }

      ++ti;
   }


   //if( !this->m_NumberOfPixelsCounted )
   //{
   //   itkExceptionMacro(<<"All the points mapped to outside of the moving image");
   //}
   //else
   {
      measure /= numberOfVoxelsInImage;
   }



   EVAL( unsigned int end_time = clock() );
//   EVAL( unsigned int end_time2 = time(0) );

   EVAL( double elapsed_time = (end_time - start_time )/(double)CLOCKS_PER_SEC );
   PRINT(" Run time for one function evaluation (using clock())    =\n"
	 << elapsed_time       << "  sec  =  "
	 << (elapsed_time/60.0) << "  min  =  "
	 << (elapsed_time/3600.0) << "  hr");
   PRINT("IJ Run_time_for_function "  << elapsed_time);
//    std::cout << " Run time for one function evaluation (using clock())    =\n"
//              << elapsed_time       << "  sec  =  ";
//    std::cout << (elapsed_time/60.0) << "  min  =  ";
//    std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;

//    double elapsed_time2 = (double)(end_time2 - start_time2 );
//    std::cout << " Run time for one function evaluation (using time(0))    =\n"
//              << elapsed_time2       << "  sec  =  ";
//    std::cout << (elapsed_time2/60.0) << "  min  =  ";
//    std::cout << (elapsed_time2/3600.0) << "  hr" << std::endl;




//    PRINTNL("IJ parameters_for_function ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(parameters[i] << " ");
//    PRINT("");

//    std::cout << "measure: " << measure << std::endl;
//    // Get the minimum of 15 and the number of parameters
//    int mini = std::min(15U,this->GetNumberOfParameters() );
//    std::cout << "parameters: ";
//    for(int i=0;i<mini;i++)
//       std::cout << parameters[i] << ", ";
//    std::cout << "\n";


   return measure;

}


/*
 * Get the Derivative Measure
 */
template < class TFixedImage, class TMovingImage>
void
ADMeanSquaresImageToImageMetric<TFixedImage,TMovingImage>
::GetDerivative( const TransformParametersType & parameters,
		 DerivativeType & derivative  ) const
{
   MeasureType measure;
   this->GetValueAndDerivative(parameters,measure,derivative);
}




/*
 * Get both the match Measure and the Derivative Measure 
 */
template <class TFixedImage, class TMovingImage> 
void
ADMeanSquaresImageToImageMetric<TFixedImage,TMovingImage>
::GetValueAndDerivative(const TransformParametersType & parameters, 
			MeasureType & measure, DerivativeType & derivative) const
{

//   EVAL( unsigned int start_time2 = time(0) );
	EVAL( unsigned int start_time = clock() );




   itkDebugMacro("GetValueAndDerivative( " << parameters << " ) ");

   FixedImageConstPointer fixedImage = this->GetFixedImage();
   MovingImageConstPointer movingImage = this->GetMovingImage();

   if( !fixedImage ) 
   {
      itkExceptionMacro( << "Fixed image has not been assigned" );
   }
   if( !movingImage ) 
   {
      itkExceptionMacro( << "Moving image has not been assigned" );
   }
   
   const unsigned int ParametersDimension = this->GetNumberOfParameters();
   const unsigned int ImageDimension = MovingImageType::ImageDimension;
   

   const unsigned int numberOfVoxelsInImage = 
      fixedImage->GetLargestPossibleRegion().GetNumberOfPixels();

   this->m_NumberOfPixelsCounted = 0;

   MeasureType measureTmp = itk::NumericTraits< MeasureType >::Zero;
   measure = itk::NumericTraits< MeasureType >::Zero;

   derivative = DerivativeType( ParametersDimension );
   derivative.Fill(itk::NumericTraits<typename DerivativeType::ValueType>::Zero);

   typename FixedImageType::IndexType index; index.Fill(0);
   typename Superclass::InputPointType inputPoint; inputPoint.Fill(0.0);
   typename Superclass::OutputPointType transformedPoint; transformedPoint.Fill(0.0);

   //-------------------------------------------------------------
   // Cast from itkTransform to itkekADTransformBase

   typedef  itkek::ADRevolveTransformBase< 
      double, ImageDimension, FixedImageType::ImageDimension>  ADTransformType;
   
   ADTransformType* adTransform =
      dynamic_cast< ADTransformType* >(this->m_Transform.GetPointer());
   
   if(!adTransform)
   {
      itkExceptionMacro(<< "Error: Could not cast to ADTransformBase");
   }

   //-------------------------------------------------------------
   typedef  itk::ImageRegionConstIteratorWithIndex<FixedImageType> FixedIteratorType;
   FixedIteratorType fixedIterator( fixedImage, this->GetFixedImageRegion() );


   //-------------------------------------------------------------

   //typename GradientInterpolatorType::Pointer gradientInterpolatorMoving = 
   //   GradientInterpolatorType::New();
   //gradientInterpolatorMoving->SetInputImage( this->m_GradientImage );

   //-------------------------------------------------------------
   //adouble parametersAD[ParametersDimension];
   adouble* parametersAD = new adouble[ParametersDimension];
   //double derivativeTmp[ParametersDimension];
   double* derivativeTmp = new double[ParametersDimension];
   short int tag = 0;
   short int keep = 1;
   int objectiveFunctionStatus = 0;
   //int stats[11] = {0};
   unsigned int counter = 0;
   unsigned int maxNumberIterPerADTape = 10000;

   adoublev transformedPointAD(ImageDimension);
   adouble movingValueAD(0.0), diffAD(0.0), valAD(0.0), measureAD(0.0);


   transformedPointAD = 0.0;

   for(unsigned int i=0;i<ParametersDimension;i++)
      parametersAD[i] = derivativeTmp[i] = 0.0;



   //-------------------------------------------------------------
   trace_on(tag,keep);

   for(unsigned int i=0;i<ParametersDimension;i++)
      parametersAD[i] <<= parameters[i];

   measureAD = 0.0;
   measureTmp = 0.0;

//   this->SetTransformParameters( parameters );
   adTransform->SetParametersAD( parametersAD );



   while(!fixedIterator.IsAtEnd())
   {

      index = fixedIterator.GetIndex();

      fixedImage->TransformIndexToPhysicalPoint( index, inputPoint );


      adTransform->TransformPointAD( inputPoint, transformedPointAD );


      //Convert adoublev back to itk Array
      for (unsigned int i=0; i < ImageDimension; i++)
      	 //transformedPoint[i] = value( transformedPointAD[i] );
	 //transformedPoint[i] = transformedPointAD[i].value();
	 transformedPoint[i] = transformedPointAD.ref(i);
  
      if( this->m_Interpolator->IsInsideBuffer( transformedPoint ) )
      {
//	  typename GradientInterpolatorType::OutputType gradientPoint = 
//	    gradientInterpolatorMoving->Evaluate( transformedPoint );

	// Get the gradient by NearestNeighboorInterpolation: 
	// which is equivalent to round up the point components.
	typedef typename  Superclass::OutputPointType OutputPointType;
	typedef typename  OutputPointType::CoordRepType CoordRepType;
	typedef itk::ContinuousIndex<CoordRepType,ImageDimension>
	  MovingImageContinuousIndexType;
    
	MovingImageContinuousIndexType tempIndex;
	this->m_MovingImage->TransformPhysicalPointToContinuousIndex( 
	   transformedPoint, tempIndex );

	typename  MovingImageType::IndexType mappedIndex; 
	for( unsigned int j = 0; j < MovingImageType::ImageDimension; j++ )
	  {
	    mappedIndex[j] = static_cast<long>( vnl_math_rnd( tempIndex[j] ) );
	  }
    
	const GradientPixelType gradientPoint = 
	  this->m_GradientImage->GetPixel( mappedIndex );




	 movingValueAD = 
	    m_ADOLCImageValue.GetImageValue( transformedPointAD,
					     gradientPoint,
					     this->m_Interpolator->Evaluate( 
						transformedPoint ));

	 const double fixedValue   = fixedIterator.Get();


	 diffAD = movingValueAD - fixedValue; 
	 measureAD += (diffAD * diffAD); 


      }
      else
      {


       //const RealType movingValue  = 0.0;
	 const RealType fixedValue   = fixedIterator.Get();
       //const RealType diff = movingValue - fixedValue; 
       //measure += diff * diff; 
	 measureAD += fixedValue * fixedValue;

      } 
      
      counter++;
      ++fixedIterator;
      this->m_NumberOfPixelsCounted++;


      if( counter == maxNumberIterPerADTape && 
	  counter != numberOfVoxelsInImage-1) 
      {
	 counter = 0;

	 // Evaluate gradient and restart AD tape. This prevents
	 // memory overflow.
	 valAD = measureAD;
	 valAD >>= measureTmp;
	 trace_off();
	 measure += measureTmp;

	 //objectiveFunctionStatus = gradient(tag,
	 //				    ParametersDimension,
	 //				    (double *)parameters.data_block(),
	 //				    derivativeTmp); 
	 //objectiveFunctionStatus = reverse(tag,1,
	 //				   ParametersDimension,0,
	 //				   1.0,
	 //				   derivativeTmp); 
	 double weight = 1.0;
	 objectiveFunctionStatus = fos_reverse(tag,
					       1,
					       ParametersDimension,
					       &weight,
					       derivativeTmp); 


	 //tapestats(tag,stats);

	 for (unsigned int i=0; i < ParametersDimension; i++)
	 {
	    derivative[i] += derivativeTmp[i];
	 }


	 trace_on(tag,keep);
	 for(unsigned int i=0;i<ParametersDimension;i++) 
	    parametersAD[i] <<= parameters[i];
	 measureAD = 0.0;
	 measureTmp = 0.0;


	 adTransform->SetParametersAD( parametersAD );
	 
      }
      
   }
    

   // Evaluate gradient.
   valAD = measureAD;
   valAD >>= measureTmp;
   trace_off();
   measure += measureTmp;

   //objectiveFunctionStatus = gradient(tag,
   //			      ParametersDimension,
   //			      (double *)parameters.data_block(),
   //			      derivativeTmp); 
   //objectiveFunctionStatus = reverse(tag,1,
   //			     ParametersDimension,0,
   //			     1.0,
   //			     derivativeTmp); 


   double weight = 1.0;
   objectiveFunctionStatus = fos_reverse(tag,
					 1,
					 ParametersDimension,
					 &weight,
					 derivativeTmp); 



   //tapestats(tag,stats);

  
   for (unsigned int i=0; i < ParametersDimension; i++)
   {
      derivative[i] += derivativeTmp[i];
   }


   //if( !this->m_NumberOfPixelsCounted )
   //{
   //itkExceptionMacro(<<"All the points mapped to outside of the moving image");
   //}
   //else
   {
      // Finally divide both the function value and the derivative by
      // the total number of voxels in the image
      measure /= numberOfVoxelsInImage;
      for (unsigned int i=0; i < ParametersDimension; i++)
      {
	 derivative[i] /= numberOfVoxelsInImage;
      }
   }


   EVAL( unsigned int end_time = clock() );
//   EVAL( unsigned int end_time2 = time(0) );


   EVAL( double elapsed_time = (end_time - start_time )/(double)CLOCKS_PER_SEC );
   PRINT(" Run time for one function evaluation and its "
	 << "gradient (using clock())    =\n"
	 << elapsed_time       << "  sec  =  "
	 << (elapsed_time/60.0) << "  min  =  "
	 << (elapsed_time/3600.0) << "  hr");
   PRINT("IJ Run_time_for_function_and_gradient " << elapsed_time);
//    std::cout << " Run time for one function evaluation and its "
//              << "gradient (using clock())    =\n"
//              << elapsed_time       << "  sec  =  ";
//    std::cout << (elapsed_time/60.0) << "  min  =  ";
//    std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;

//    double elapsed_time2 = (double)(end_time2 - start_time2 );
//    std::cout << " Run time for one function evaluation and its "
//              << "gradient (using time(0))    =\n"
//              << elapsed_time2       << "  sec  =  ";
//    std::cout << (elapsed_time2/60.0) << "  min  =  ";
//    std::cout << (elapsed_time2/3600.0) << "  hr" << std::endl;

//    PRINTNL("IJ parameters_for_function_and_gradient ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(parameters[i] << " ");
//    PRINT("");
//    PRINTNL("IJ derivative_for_function_and_gradient ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(derivative[i] << " ");
//    PRINT("");


//    std::cout << "measure: " << measure << std::endl;
//    // Get the minimum of 15 and the number of parameters
//    int mini = std::min(15U,this->GetNumberOfParameters() );
//    std::cout << "parameters: ";
//    for(int i=0;i<mini;i++)
//       std::cout << parameters[i] << ", ";
//    std::cout << "\n";
//    std::cout << "derivative: ";
//    for(int i=0;i<mini;i++)
//       std::cout << derivative[i] << ", ";
//    std::cout << "\n";


   delete [] parametersAD;
   delete [] derivativeTmp;

}




}



#endif
