#ifndef __itkekADRevolveMattesMutualInformationImageToImageMetric_h
#define __itkekADRevolveMattesMutualInformationImageToImageMetric_h

#include "itkekADRevolveImageToImageMetricBase.h"

#include "itkPoint.h"
#include "itkVector.h"
#include "itkPointSet.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkCovariantVector.h"
#include "itkImage.h"
#include "itkVectorLinearInterpolateImageFunction.h"
#include "itkAffineTransform.h"
#include "itkBSplineKernelFunction.h"
#include "itkBSplineDerivativeKernelFunction.h"
#include "itkCentralDifferenceImageFunction.h"
#include "vnl/vnl_matrix.h"
#include "vnl/vnl_matrix_fixed.h"

#include <vector>




namespace itkek
{


/**
   This class represents the objective function used for non-rigid
   registration. 
 */
class ITK_EXPORT ADRevolveMattesMutualInformationImageToImageMetric : 
	 public ADRevolveImageToImageMetricBase<itk::Image<float, 3u>,
						itk::Image<float, 3u> >
{

   public:


      typedef ADRevolveMattesMutualInformationImageToImageMetric    Self;
      typedef ADRevolveImageToImageMetricBase<itk::Image<float, 3u>,
					      itk::Image<float, 3u> > Superclass;
      
      typedef itk::SmartPointer<Self>         Pointer;
      typedef itk::SmartPointer<const Self>   ConstPointer;

      /** Method for creation through the object factory. */
      itkNewMacro(Self);

      /** Run-time type information (and related methods). */
      itkTypeMacro(ADRevolveMattesMutualInformationImageToImageMetric,
		   ADRevolveImageToImageMetricBase);


      typedef itk::Point<double,3> PointType;
      typedef std::vector<PointType> PointsContainer;
      typedef itk::Vector<double,3> VectorType;


      typedef itk::Image< float,3 > MovingImageType ;
      typedef MovingImageType::ConstPointer MovingImagePointer;
      typedef itk::Image< float,3 > FixedImageType ;
      typedef FixedImageType::ConstPointer FixedImagePointer;

      typedef itk::LinearInterpolateImageFunction<
	 MovingImageType, double > 	InterpolatorType;
      typedef InterpolatorType::Pointer 	MovingImageInterpolatorPointer;

      typedef itk::AffineTransform< double, 3 > AffineTransformType;
      typedef AffineTransformType::Pointer AffineTransformPointer;

      typedef enum { 
	 OptimizeOnDisplacements, OptimizeOnCoefficients 
      } WhatToOptimizeOnType;


      void SetNumberOfHistogramBins(unsigned long bins) const
      {s_NumberOfHistogramBins = bins;}

      void UseAllPixelsOn() const {s_UseAllPixels = true;}
      void UseAllPixelsOff() const {s_UseAllPixels = false;}

   protected:

      ADRevolveMattesMutualInformationImageToImageMetric() {}
      virtual ~ADRevolveMattesMutualInformationImageToImageMetric() {}


   private:


      typedef PointType InputPointType;
      typedef PointType OutputPointType;
      typedef itk::Vector<double,3> InputVectorType;
      typedef itk::Vector<double,3> OutputVectorType;



      typedef adoublev ADVectorType;
      typedef adoublem ADMatrixType;


      typedef itk::CovariantVector< double,3 >  GradientPixelType;
      typedef itk::Image< GradientPixelType,3 > GradientImageType ;
      typedef itk::SmartPointer< GradientImageType > GradientImagePointer;


      typedef  itk::VectorLinearInterpolateImageFunction<
         GradientImageType, double >    GradientInterpolatorType;
      typedef GradientInterpolatorType::Pointer GradientInterpolatorPointer;





      //These variables are static since they don't change 
      // throughout the entire function evaluation

      
      // A fixed image spatial sample consists of the fixed domain point 
      // and the fixed image value at that point. 
      class FixedImageSpatialSample
      {
	 public:
	    FixedImageSpatialSample():FixedImageValue(0.0)
	    { FixedImagePointValue.Fill(0.0); }
	    ~FixedImageSpatialSample() {};

	    FixedImageType::PointType           FixedImagePointValue;
	    double                        FixedImageValue;
	    unsigned int                  FixedImageParzenWindowIndex;
      };

      typedef double PDFValueType;
      typedef itk::Image< PDFValueType, 2 >  JointPDFType;
//      typedef Image<PDFValueType,3> JointPDFDerivativesType;
      typedef JointPDFType::IndexType                JointPDFIndexType;
      typedef JointPDFType::PixelType                JointPDFValueType;
      typedef JointPDFType::RegionType              JointPDFRegionType;
      typedef JointPDFType::SizeType                JointPDFSizeType;
//      typedef JointPDFDerivativesType::IndexType    JointPDFDerivativesIndexType;
//      typedef JointPDFDerivativesType::PixelType    JointPDFDerivativesValueType;
//      typedef JointPDFDerivativesType::RegionType  JointPDFDerivativesRegionType;
//      typedef JointPDFDerivativesType::SizeType      JointPDFDerivativesSizeType;
      typedef itk::Image<adouble,2> JointPDFTypeAD;



      /** FixedImageSpatialSample typedef support. */
      typedef std::vector<FixedImageSpatialSample>
                 FixedImageSpatialSampleContainer;

      /** The marginal PDFs are stored as std::vector. */
      //typedef float PDFValueType;
      typedef std::vector<PDFValueType> MarginalPDFType;

      /** Typedefs for BSpline kernel and derivative functions. */
      typedef itk::BSplineKernelFunction<3> CubicBSplineFunctionType;
      typedef itk::BSplineDerivativeKernelFunction<3>
      CubicBSplineDerivativeFunctionType;

      /** Typedefs for using central difference calculator. */
      typedef itk::CentralDifferenceImageFunction<MovingImageType,
                           double> DerivativeFunctionType;

      /** Pointer to central difference calculator. */
      static DerivativeFunctionType::Pointer s_DerivativeCalculator;
       

      /** Cubic BSpline kernel for computing Parzen histograms. */
      static CubicBSplineFunctionType::Pointer s_CubicBSplineKernel;
      static CubicBSplineDerivativeFunctionType::Pointer
      s_CubicBSplineDerivativeKernel;


      // Note we only use this for its meta-information
      // about the joint pdf which is constant throughout the
      // entire function evaluation. The data itself changes
      // and is stored in the base class (m_Adoubles)
      /** The joint PDF and PDF derivatives. */
      // static JointPDFDerivativesType::Pointer m_JointPDFDerivatives;
      static JointPDFType::Pointer s_JointPDF;

      /** Container to store a set of points and fixed image values. */
      static FixedImageSpatialSampleContainer    s_FixedImageSamples;

      /** The fixed image marginal PDF. */
      static MarginalPDFType s_FixedImageMarginalPDF;
      static MarginalPDFType s_MovingImageMarginalPDF;

      static bool s_UseAllPixels;
      static unsigned long s_NumberOfSpatialSamples;
      static unsigned long s_NumberOfParameters;
      static unsigned long s_NumberOfHistogramBins;
      static double s_MovingImageNormalizedMin;
      static double s_FixedImageNormalizedMin;
      static double s_MovingImageTrueMin;
      static double s_MovingImageTrueMax;
      static double s_FixedImageBinSize;
      static double s_MovingImageBinSize;

      static JointPDFTypeAD::Pointer s_JointPDFAD;



      // The main function.
      virtual void Function(Label& label);

      virtual adolc::ObjectiveFunctionBase* Create() const;

      virtual LabelsContainerType GetLabels() const;

      virtual void CopyData(const ObjectiveFunctionBase* rhs);


      
      void Initialize() throw ( itk::ExceptionObject );

      /** Precompute fixed image parzen window indices. */
      void ComputeFixedImageParzenWindowIndices( 
                           FixedImageSpatialSampleContainer& samples );
      
      /** Uniformly select a sample set from the fixed image domain. */        
      void SampleFixedImageDomain( 
	 FixedImageSpatialSampleContainer& samples);

      void SampleFullFixedImageDomain( 
	 FixedImageSpatialSampleContainer& samples );

      void SampleFixedImageDomainForJacobian() const;

      void ReinitializeSeed();

      void ReinitializeSeed(int seed);

      void ComputePDFDerivatives( 
	 unsigned int sampleNumber, 
	 int pdfMovingIndex,
	 const GradientInterpolatorType::OutputType& movingImageGradientValue,
	 double cubicBSplineDerivativeValue ) const;




      void MI(Label& label);


      /** Transform a point from FixedImage domain to MovingImage domain.
       * This function also checks if mapped point is within support
       * region. */
      void TransformPoint( unsigned int sampleNumber, 
			   ADVectorType& transformedPoint, 
			   bool& sampleOk,
			   adouble& movingImageValue) const;


   private:
      ADRevolveMattesMutualInformationImageToImageMetric(
	 const ADRevolveMattesMutualInformationImageToImageMetric& rhs); // not implemented
      ADRevolveMattesMutualInformationImageToImageMetric& operator=(
	 const ADRevolveMattesMutualInformationImageToImageMetric& rhs); //not implemented



      //Local data
      int m_Ints[10];

};

}


#endif
