#ifndef __itkekMeanSquaresImageToImageMetric_h
#define __itkekMeanSquaresImageToImageMetric_h

#include "itkImageToImageMetric.h"


namespace itkek
{

//  using namespace itk;

/** \class MeanSquaresImageToImageMetric
 * \brief Computes similarity between two objects to be registered
 *
 * This class is almost identical to the itk version of this class except for
 * the following changes:
 * 
 * When computing the match metric, if the point is mapped to outside the
 * image region it is still given a value of zero, rather than being entirely
 * exclude from the sum. This prevents the exception of "all points mapped to
 * outside of the moving image" from being thrown.
 * 
 *
 * It is in the namespace itkek not itk.
 * 
 * 
 * \ingroup RegistrationMetrics
 */

template < typename TFixedImage, typename TMovingImage>
class ITK_EXPORT MeanSquaresImageToImageMetric : 
 public itk::ImageToImageMetric< TFixedImage, TMovingImage >
{
public:

  /** Standard class typedefs. */
  typedef MeanSquaresImageToImageMetric    Self;
  typedef itk::ImageToImageMetric< TFixedImage, TMovingImage > Superclass;

  typedef itk::SmartPointer<Self>         Pointer;
  typedef itk::SmartPointer<const Self>   ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);
 
  /** Run-time type information (and related methods). */
  itkTypeMacro(MeanSquaresImageToImageMetric, itk::ImageToImageMetric);

 
  /** Types transferred from the base class */
  typedef typename Superclass::RealType                 RealType;
  typedef typename Superclass::TransformType            TransformType;
  typedef typename Superclass::TransformPointer         TransformPointer;
  typedef typename Superclass::TransformParametersType  TransformParametersType;
  typedef typename Superclass::TransformJacobianType    TransformJacobianType;
  typedef typename Superclass::GradientPixelType        GradientPixelType;

  typedef typename Superclass::MeasureType              MeasureType;
  typedef typename Superclass::DerivativeType           DerivativeType;
  typedef typename Superclass::FixedImageType           FixedImageType;
  typedef typename Superclass::MovingImageType          MovingImageType;
  typedef typename Superclass::FixedImageConstPointer   FixedImageConstPointer;
  typedef typename Superclass::MovingImageConstPointer  MovingImageConstPointer;


  /** Get the derivatives of the match measure. */
  void GetDerivative( const TransformParametersType & parameters,
                      DerivativeType & derivative ) const;

  /**  Get the value for single valued optimizers. */
  MeasureType GetValue( const TransformParametersType & parameters ) const;

  /**  Get value and derivatives for multiple valued optimizers. */
  void GetValueAndDerivative( const TransformParametersType & parameters,
                              MeasureType& Value, DerivativeType& Derivative ) const;

protected:
  MeanSquaresImageToImageMetric();
  virtual ~MeanSquaresImageToImageMetric() {};

private:
  MeanSquaresImageToImageMetric(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

};

} // end namespace itkek

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkekMeanSquaresImageToImageMetric.txx"
#endif

#endif



