#ifndef _itkekMeanSquaresImageToImageMetric_cxx
#define _itkekMeanSquaresImageToImageMetric_cxx

#include "itkekMeanSquaresImageToImageMetric.h"
#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkekMacro.h"

#include <time.h>

namespace itkek
{

/*
 * Constructor
 */
template <  typename TFixedImage, typename TMovingImage >
MeanSquaresImageToImageMetric< TFixedImage, TMovingImage >
::MeanSquaresImageToImageMetric()
{
//  itkDebugMacro("Constructor");
}

/*
 * Get the match Measure
 */
template <  typename TFixedImage, typename TMovingImage >
typename MeanSquaresImageToImageMetric< TFixedImage, TMovingImage >::MeasureType
MeanSquaresImageToImageMetric< TFixedImage, TMovingImage >
::GetValue( const TransformParametersType & parameters ) const
{

//   EVAL( unsigned int start_time2 = time(0) );
	EVAL( unsigned int start_time = clock() );



//  itkDebugMacro("GetValue( " << parameters << " ) ");

  FixedImageConstPointer fixedImage = this->GetFixedImage();

  if( !fixedImage ) 
    {
    itkExceptionMacro( << "Fixed image has not been assigned" );
    }

  typedef  itk::ImageRegionConstIteratorWithIndex<FixedImageType> FixedIteratorType;


  FixedIteratorType ti( fixedImage, this->GetFixedImageRegion() );

  typename FixedImageType::IndexType index;

  MeasureType measure = itk::NumericTraits< MeasureType >::Zero;

  this->m_NumberOfPixelsCounted = 0;


  this->SetTransformParameters( parameters );


  while(!ti.IsAtEnd())
    {

    index = ti.GetIndex();
    
    typename  Superclass::InputPointType inputPoint;
    fixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

     typename Superclass::OutputPointType transformedPoint = 
      this->m_Transform->TransformPoint( inputPoint );

    if( this->m_Interpolator->IsInsideBuffer( transformedPoint ) )
      {
      const RealType movingValue  = this->m_Interpolator->Evaluate( transformedPoint );
      const RealType fixedValue   = ti.Get();
      this->m_NumberOfPixelsCounted++;
      const RealType diff = movingValue - fixedValue; 
      measure += diff * diff; 
      }
    else
      {
      //const RealType movingValue  = 0.0;
      const RealType fixedValue   = ti.Get();
      this->m_NumberOfPixelsCounted++;
      //const RealType diff = movingValue - fixedValue; 
      //measure += diff * diff; 
      measure += fixedValue * fixedValue;
      }

    ++ti;
    }

  if( !this->m_NumberOfPixelsCounted )
    {
    itkExceptionMacro(<<"All the points mapped to outside of the moving image");
    }
  else
    {
    measure /= this->m_NumberOfPixelsCounted;
    }




  EVAL( unsigned int end_time = clock() );
//   EVAL( unsigned int end_time2 = time(0) );

  EVAL( double elapsed_time = (end_time - start_time )/(double)CLOCKS_PER_SEC );
   PRINT(" Run time for one function evaluation (using clock())    =\n"
	 << elapsed_time       << "  sec  =  "
	 << (elapsed_time/60.0) << "  min  =  "
	 << (elapsed_time/3600.0) << "  hr");
   PRINT("IJ Run_time_for_function "  << elapsed_time);
//    std::cout << " Run time for one function evaluation (using clock())    =\n"
//              << elapsed_time       << "  sec  =  ";
//    std::cout << (elapsed_time/60.0) << "  min  =  ";
//    std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;

//    double elapsed_time2 = (double)(end_time2 - start_time2 );
//    std::cout << " Run time for one function evaluation (using time(0))    =\n"
//              << elapsed_time2       << "  sec  =  ";
//    std::cout << (elapsed_time2/60.0) << "  min  =  ";
//    std::cout << (elapsed_time2/3600.0) << "  hr" << std::endl;

//    PRINTNL("IJ parameters_for_function ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(parameters[i] << " ");
//    PRINT("");

/*   std::cout << "measure: " << measure << std::endl;
   // Get the minimum of 15 and the number of parameters
   int mini = std::min(15U,this->GetNumberOfParameters() );
   std::cout << "parameters: ";
   for(int i=0;i<mini;i++)
      std::cout << parameters[i] << ", ";
   std::cout << "\n";*/

  return measure;

}





/*
 * Get the Derivative Measure
 */
template <  typename TFixedImage, typename TMovingImage >
void
MeanSquaresImageToImageMetric< TFixedImage, TMovingImage >
::GetDerivative( const TransformParametersType & parameters,
                 DerivativeType & derivative  ) const
{

  MeasureType value;
  // call the combined version
  this->GetValueAndDerivative( parameters, value, derivative );

/*
//  itkDebugMacro("GetDerivative( " << parameters << " ) ");
  
  if( !this->m_GradientImage )
    {
    itkExceptionMacro(<<"The gradient image is null, maybe you forgot to call Initialize()");
    }

  FixedImageConstPointer fixedImage = this->GetFixedImage();

  if( !fixedImage ) 
    {
    itkExceptionMacro( << "Fixed image has not been assigned" );
    }

  const unsigned int ImageDimension = FixedImageType::ImageDimension;


  typedef  itk::ImageRegionConstIteratorWithIndex<
    FixedImageType> FixedIteratorType;

  typedef  itk::ImageRegionConstIteratorWithIndex<
   typename   Superclass::GradientImageType> GradientIteratorType;


  FixedIteratorType ti( fixedImage, this->GetFixedImageRegion() );

  typename  FixedImageType::IndexType index;

  this->m_NumberOfPixelsCounted = 0;

  this->SetTransformParameters( parameters );

  const unsigned int ParametersDimension = this->GetNumberOfParameters();
  derivative = DerivativeType( ParametersDimension );
  derivative.Fill( itk::NumericTraits< typename DerivativeType::ValueType>::Zero );

  ti.GoToBegin();

  while(!ti.IsAtEnd())
    {

    index = ti.GetIndex();
    
    typename Superclass::InputPointType inputPoint;
    fixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

    typename  Superclass::OutputPointType transformedPoint = 
      this->m_Transform->TransformPoint( inputPoint );

    if( this->m_Interpolator->IsInsideBuffer( transformedPoint ) )
      {
	const RealType movingValue  = this->m_Interpolator->Evaluate( transformedPoint );

	const TransformJacobianType & jacobian =
	  this->m_Transform->GetJacobian( inputPoint ); 

      
	const RealType fixedValue     = ti.Value();
	this->m_NumberOfPixelsCounted++;
	const RealType diff = movingValue - fixedValue; 
    
	// Get the gradient by NearestNeighboorInterpolation: 
	// which is equivalent to round up the point components.
	typedef typename  Superclass::OutputPointType OutputPointType;
	typedef typename  OutputPointType::CoordRepType CoordRepType;
	typedef itk::ContinuousIndex<CoordRepType,MovingImageType::ImageDimension>
	  MovingImageContinuousIndexType;
    
	MovingImageContinuousIndexType tempIndex;
	this->m_MovingImage->TransformPhysicalPointToContinuousIndex( transformedPoint, tempIndex );
    
	typename  MovingImageType::IndexType mappedIndex; 
	for( unsigned int j = 0; j < MovingImageType::ImageDimension; j++ )
	  {
	    mappedIndex[j] = static_cast<long>( vnl_math_rnd( tempIndex[j] ) );
	  }
    
	const GradientPixelType gradient = 
	  this->m_GradientImage->GetPixel( mappedIndex );
    
	for(unsigned int par=0; par<ParametersDimension; par++)
	  {
	    RealType sum = itk::NumericTraits< RealType >::Zero;
	    for(unsigned int dim=0; dim<ImageDimension; dim++)
	      {
		sum += 2.0 * diff * jacobian( dim, par ) * gradient[dim];
	      }
	    derivative[par] += sum;
	  }
      }    
    
    ++ti;
    }
  
  if( !this->m_NumberOfPixelsCounted )
    {
      itkExceptionMacro(<<"All the points mapped to outside of the moving image");
    }
  else
    {
      for(unsigned int i=0; i<ParametersDimension; i++)
	{
	  derivative[i] /= this->m_NumberOfPixelsCounted;
	}
    }
*/  
}


/*
 * Get both the match Measure and theDerivative Measure 
 */
template <  typename TFixedImage, typename TMovingImage >
void
MeanSquaresImageToImageMetric< TFixedImage, TMovingImage >
::GetValueAndDerivative(const TransformParametersType & parameters, 
                        MeasureType & value, DerivativeType  & derivative) const
{

//   EVAL( unsigned int start_time2 = time(0) );
	EVAL( unsigned int start_time = clock() );


//  itkDebugMacro("GetValueAndDerivative( " << parameters << " ) ");

  if( !this->m_GradientImage )
    {
    itkExceptionMacro(<<"The gradient image is null, maybe you forgot to call Initialize()");
    }

  FixedImageConstPointer fixedImage = this->GetFixedImage();

  if( !fixedImage ) 
    {
    itkExceptionMacro( << "Fixed image has not been assigned" );
    }

  const unsigned int ImageDimension = FixedImageType::ImageDimension;

  typedef  itk::ImageRegionConstIteratorWithIndex<
    FixedImageType> FixedIteratorType;

  typedef  itk::ImageRegionConstIteratorWithIndex<
    typename  Superclass::GradientImageType> GradientIteratorType;


  FixedIteratorType ti( fixedImage, this->GetFixedImageRegion() );

  typename  FixedImageType::IndexType index;

  MeasureType measure = itk::NumericTraits< MeasureType >::Zero;

  this->m_NumberOfPixelsCounted = 0;

  this->SetTransformParameters( parameters );

  const unsigned int ParametersDimension = this->GetNumberOfParameters();

  derivative = DerivativeType( ParametersDimension );
  derivative.Fill( itk::NumericTraits< typename DerivativeType::ValueType>::Zero );

  ti.GoToBegin();


  while(!ti.IsAtEnd())
    {
/*
      if(tt++ != 50006)
      {
	 ++ti;
	 continue;
      }
*/

    index = ti.GetIndex();

          
     typename  Superclass::InputPointType inputPoint;
    fixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

    typename  Superclass::OutputPointType transformedPoint = 
      this->m_Transform->TransformPoint( inputPoint );

    if( this->m_Interpolator->IsInsideBuffer( transformedPoint ) )
      {
	const RealType movingValue = this->m_Interpolator->Evaluate( transformedPoint );

	const TransformJacobianType & jacobian =
	  this->m_Transform->GetJacobian( inputPoint ); 
      
	const RealType fixedValue     = ti.Value();
	this->m_NumberOfPixelsCounted++;
    
	const RealType diff = movingValue - fixedValue; 
    
	measure += diff * diff;
    
	// Get the gradient by NearestNeighboorInterpolation: 
	// which is equivalent to round up the point components.
	typedef typename  Superclass::OutputPointType OutputPointType;
	typedef typename  OutputPointType::CoordRepType CoordRepType;
	typedef itk::ContinuousIndex<CoordRepType,MovingImageType::ImageDimension>
	  MovingImageContinuousIndexType;
    
	MovingImageContinuousIndexType tempIndex;
	this->m_MovingImage->TransformPhysicalPointToContinuousIndex( transformedPoint, tempIndex );

	typename  MovingImageType::IndexType mappedIndex; 
	for( unsigned int j = 0; j < MovingImageType::ImageDimension; j++ )
	  {
	    mappedIndex[j] = static_cast<long>( vnl_math_rnd( tempIndex[j] ) );
	  }
    
	const GradientPixelType gradient = 
	  this->m_GradientImage->GetPixel( mappedIndex );
/*
	if(gradient[0] != 0.0)
	{
	   std::cout << inputPoint << "\n";
	   std::cout << gradient << "\n";
	   std::cout << tt << "\n";
	   std::cout << index << "\n";
	   std::cout << jacobian << "\n";
	}
*/
	for(unsigned int par=0; par<ParametersDimension; par++)
	  {
	    RealType sum = itk::NumericTraits< RealType >::Zero;
	    for(unsigned int dim=0; dim<ImageDimension; dim++)
	      {
		 sum += 2.0 * diff * jacobian( dim, par ) * gradient[dim];
	      }
	    derivative[par] += sum;
	  }
      }      
    else
      {
      // If transformedPoint is not inside the image buffer
      // then the derivative is zero so just update the metric value.

      //const RealType movingValue  = 0.0;
      const RealType fixedValue   = ti.Get();
      this->m_NumberOfPixelsCounted++;
      //const RealType diff = movingValue - fixedValue; 
      //measure += diff * diff; 
      measure += fixedValue * fixedValue;
      }

    ++ti;
    }

  if( !this->m_NumberOfPixelsCounted )
    {
    itkExceptionMacro(<<"All the points mapped to outside of the moving image");
    }
  else
    {
    for(unsigned int i=0; i<ParametersDimension; i++)
      {
      derivative[i] /= this->m_NumberOfPixelsCounted;
      }
    measure /= this->m_NumberOfPixelsCounted;
    }

  value = measure;



  EVAL( unsigned int end_time = clock() );
//   EVAL( unsigned int end_time2 = time(0) );


  EVAL( double elapsed_time = (end_time - start_time )/(double)CLOCKS_PER_SEC );
   PRINT(" Run time for one function evaluation and its "
	 << "gradient (using clock())    =\n"
	 << elapsed_time       << "  sec  =  "
	 << (elapsed_time/60.0) << "  min  =  "
	 << (elapsed_time/3600.0) << "  hr");
   PRINT("IJ Run_time_for_function_and_gradient " << elapsed_time);
//    std::cout << " Run time for one function evaluation and its "
//              << "gradient (using clock())    =\n"
//              << elapsed_time       << "  sec  =  ";
//    std::cout << (elapsed_time/60.0) << "  min  =  ";
//   std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;

//    double elapsed_time2 = (double)(end_time2 - start_time2 );
//    std::cout << " Run time for one function evaluation and its "
//              << "gradient (using time(0))    =\n"
//              << elapsed_time2       << "  sec  =  ";
//    std::cout << (elapsed_time2/60.0) << "  min  =  ";
//    std::cout << (elapsed_time2/3600.0) << "  hr" << std::endl;


//    PRINTNL("IJ parameters_for_function_and_gradient ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(parameters[i] << " ");
//    PRINT("");
//    PRINTNL("IJ derivative_for_function_and_gradient ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(derivative[i] << " ");
//    PRINT("");
/*   std::cout << "measure: " << measure << std::endl;
   // Get the minimum of 15 and the number of parameters
   int mini = std::min(15U,this->GetNumberOfParameters() );
   std::cout << "parameters: ";
   for(int i=0;i<mini;i++)
      std::cout << parameters[i] << ", ";
   std::cout << "\n";
   std::cout << "derivative: ";
   for(int i=0;i<mini;i++)
      std::cout << derivative[i] << ", ";
   std::cout << "\n";*/

}

} // end namespace itk


#endif
