

#include "itkCastImageFilter.h"
#include "itkCenteredTransformInitializer.h"
#include "itkCommand.h"
#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkLBFGSBOptimizer.h"
#include "itkRegularStepGradientDescentOptimizer.h"
#include "itkMultiResolutionImageRegistrationMethod.h"
#include "vnl/vnl_sample.h"

#include "itkekMacro.h"
#include "itkekMeanSquaresImageToImageMetric.h"
#include "itkekADRevolveMatrixOffsetTransformBase.h"
#include "itkekADMeanSquaresImageToImageMetric.h"
#include "itkekADMattesMutualInformationImageToImageMetric.h"
#include "itkekADRevolveBSplineDeformableTransform.h"
#include "itkekADRevolveEuler3DTransform.h"
#include "itkekMattesMutualInformationImageToImageMetric.h"
#include "itkekADRevolveMeanSquaresImageToImageMetric.h"
#include "itkekADRevolveMattesMutualInformationImageToImageMetric.h"
#include "itkekLoadSaveAffineTransform.h"
#include "itkekADRevolveRBFTransform.h"



#include <fstream>
#include <ctime>
#include "getopt.h"


//  The following section of code implements a Command observer
//  that will monitor the evolution of the optimizer.
class CommandIterationUpdate : public itk::Command 
{
public:
	typedef  CommandIterationUpdate   Self;
	typedef  itk::Command             Superclass;
	typedef  itk::SmartPointer<Self>  Pointer;
	itkNewMacro( Self );
protected:
	CommandIterationUpdate() {};
public:
	typedef   itk::RegularStepGradientDescentOptimizer     OptimizerType;
	typedef   itk::LBFGSBOptimizer                         OptimizerType2;
	
	void Execute(itk::Object * object, const itk::EventObject & event)
	{
		
		if( typeid( event ) == typeid( itk::IterationEvent ) )
		{
			if(typeid(*object) == typeid(OptimizerType))
			{
				EVAL( OptimizerType* optimizer = 
					  dynamic_cast< OptimizerType* >( object ) );
				
				PRINT("\nIteration:         " 
					  << optimizer->GetCurrentIteration() 
					  << "\nMetric value:      " 
					  << optimizer->GetValue());
				PRINTNL("IJ parameters_for_function ");
				EVAL(for(unsigned int i=0;i<optimizer->GetCurrentPosition().size();i++))
					PRINTNL(optimizer->GetCurrentPosition()[i] << " ");
				PRINT("\n");
				/*std::cout << "\nIteration:         " 
				  << optimizer->GetCurrentIteration() 
				  << "\nMetric value:      " 
				  << optimizer->GetValue()
				  << "\nCurrent position:  " 
				  << optimizer->GetCurrentPosition() 
				  << "\n" << std::endl;*/
				
			}
			else if(typeid(*object) == typeid(OptimizerType2))
			{
				EVAL( OptimizerType2* optimizer = 
					  dynamic_cast< OptimizerType2* >( object ) );
				
				PRINT("\nIteration:         " 
					  << optimizer->GetCurrentIteration() 
					  << "\nMetric value:      " 
					  << optimizer->GetValue());
				PRINTNL("IJ parameters_for_function ");
				EVAL(for(unsigned int i=0;i<optimizer->GetCurrentPosition().size();i++))
					PRINTNL(optimizer->GetCurrentPosition()[i] << " ");
				PRINT("\n");
				/*std::cout << "\nIteration:         " 
				  << optimizer->GetCurrentIteration() 
				  << "\nMetric value:      " 
				  << optimizer->GetValue()
				  << "\nCurrent position:  " 
				  << optimizer->GetCurrentPosition() 
				  << "\n" << std::endl;*/
			}
		}
	}
	

	void Execute(const itk::Object *caller, const itk::EventObject & event)
	{
		Execute( const_cast<itk::Object *>(caller), event);
	}
	
	
};




//  The following section of code implements a Command observer
//  that will control the modification of optimizer parameters
//  at every change of resolution level.
template <typename TRegistration>
class RegistrationInterfaceCommand : public itk::Command 
{
public:
	typedef  RegistrationInterfaceCommand   Self;
	typedef  itk::Command                   Superclass;
	typedef  itk::SmartPointer<Self>        Pointer;
	itkNewMacro( Self );
protected:
	RegistrationInterfaceCommand() {};
public:
	typedef   TRegistration                              RegistrationType;
	
      
	typedef   RegistrationType *                         RegistrationPointer;
	void Execute(itk::Object * object, const itk::EventObject & event)
	{
		
		RegistrationPointer registration =
			dynamic_cast<RegistrationPointer>( object );
		
		if(!registration)
			return;
		
		int numberLevels = registration->GetNumberOfLevels();
		int currentLevel = registration->GetCurrentLevel();
		
		if ( numberLevels > 1 )
		{
			
			PRINT("\nStarting resolution level " 
				  << currentLevel
				  << " of registration...\n");
			/*std::cout << "\nStarting resolution level " 
		      << currentLevel
		      << " of registration...\n"
		      << std::endl;*/
			
			typedef itkek::MattesMutualInformationImageToImageMetric< 
			typename RegistrationType::FixedImageType, 
				typename RegistrationType::MovingImageType >    MetricType;
			
			if(typeid(*registration->GetMetric()) == typeid(MetricType))
			{
				MetricType* metric = 
					dynamic_cast< MetricType* >( registration->GetMetric() );
				
				
				const unsigned int numberOfPixels =
					registration->GetFixedImagePyramid()->GetOutput(currentLevel)->
					GetLargestPossibleRegion().GetNumberOfPixels();
				
				const unsigned int numberOfSamples = numberOfPixels / 10;
				metric->SetNumberOfSpatialSamples( numberOfSamples );
				
				PRINT("setting number of samples to: " 
					  << numberOfSamples);
				/*std::cout << "setting number of samples to: " 
				  << numberOfSamples << std::endl;*/
			}
			
			
		}
		
		
		
		// Since the images of highest resolution level are actually
		// different than the original fixed and moving images (since they
		// were blurred), we replace them with the original images instead.
		if(currentLevel == numberLevels-1)
		{
			
			typedef typename RegistrationType::FixedImageType FixedImageType;
			typedef typename RegistrationType::MovingImageType MovingImageType;
			
			registration->GetFixedImagePyramid()->GetOutputs()[currentLevel] =
				const_cast<FixedImageType*>(registration->GetFixedImage());
			
			registration->GetMovingImagePyramid()->GetOutputs()[currentLevel] =
				const_cast<MovingImageType*>(registration->GetMovingImage());
			
			
		}
		
	}
	void Execute(const itk::Object * object, const itk::EventObject & event)
	{ return; }
	
};



template <class PointSetSmartPointer>
static bool LoadPointSet(const char* const  filename,
						 const PointSetSmartPointer pointset)
{
	
	std::ifstream pointsfile(filename);
	if(!pointsfile)
	{
		itkGenericExceptionMacro(<< "Error: cannot open file");
	}
	
	
	if(pointset.IsNull())
	{
		itkGenericExceptionMacro(<< "Error: pointer has not been initialized");
	}


	typedef typename PointSetSmartPointer::ObjectType PointSetType;
	typename PointSetType::PointType p;

	int i=0;
	while( pointsfile >> p[0] >> p[1] >> p[2])
	{
		pointset->SetPoint(i,p);
		i++;
	}
	
	pointsfile.close();
	
	return 0;
	
}




static void usage()
{
	std::cout << "Usage: Register fixedImageFile movingImageFile [OPTIONS]\n\n\
This program registers images.\n\n\
 Possible options:\n\
 -h, --help \n\
       display this help\n\
 -i INTEGER, --max-iter=INTEGER \n\
       maximum number of iterations for optimizer (default: 25)\n\
 -m INTEGER, --metric-type=INTEGER \n\
       metric type (default: 1) where INTEGER can be \n\
       1 - sum of squared differences with automatic differentiation and\n\
           checkpointing\n\
       2 - mutual information with automatic differentiation and checkpointing\n\
       3 - sum of squared differences with automatic differentiation but no\n\
           checkpointing\n\
       4 - broken, do not use\n\
       5 - sum of squared differences without automatic differentiation or\n\
           checkpointing\n\
       6 - mutual information without automatic differentiation or checkpointing\n\
 -t INTEGER, --transform-type=INTEGER \n\
       transform type (default: 1) where INTEGER can be\n\
       1 - rigid\n\
       2 - affine\n\
       3 - B-Spline\n\
       4 - thin plate spline\n\
 -z INTEGER, --optimizer-type=INTEGER \n\
       optimizer type (default: 1) where INTEGER can be\n\
       1 - RegularStepGradientDescentOptimizer\n\
       2 - LBFGSBOptimizer\n\
 -p <filename>, --parameters-file <filename>\n\
       File to store transformation parameters (default: ./transform_parameters)\n\
 -r INTEGER, --resolution-levels=INTEGER \n\
       number of resolution levels (default: 1)\n\
" << std::endl;

	exit(1);
}


int main( int argc, char **argv )
{

	bool        arg_affine_transform=0;
	bool        arg_testing_only = 0;
	int         arg_metric_type=1;
	int         arg_transform_type=1;
	int         arg_optimizer_type=1;
	int         arg_max_iter=25;
	int         arg_resolution_levels=1;
	int         arg_histogram_bins=50;
	double      arg_grid_spacing=10.0;
	const char* arg_parameters_file="transform_parameters";
	const char* arg_control_points_file=0;
	bool        arg_constrained_points=0;
	const char* arg_constrained_points_file="constrained_points.txt";
	bool        arg_hard_constrained_points=0;
	const char* arg_hard_constrained_points_file="hard_constrained_points.txt";
	const char* arg_affine_transform_file=0;
	double      arg_tau = 0.0;



	optind = 1;
	while (1)
	{
		static const struct option long_options [] = {
			{"help",                          no_argument, 0, 'h'},
			{"metric-type",             required_argument, 0, 'm'},
			{"transform-type",          required_argument, 0, 't'},
			{"optimizer-type",          required_argument, 0, 'z'},
			{"parameters-file",         required_argument, 0, 'p'},
			{"max-iter",                required_argument, 0, 'i'},
			{"resolution-levels",       required_argument, 0, 'r'},
			{"histogram-bins",          required_argument, 0, 'b'},
			{"grid-spacing",            required_argument, 0, 's'},
			{"control-points-file",     required_argument, 0, 'c'},
			{"constrained-points-file", required_argument, 0, 'n'},
			{"affine-transform",        required_argument, 0, 'a'},
			{"tau",                     required_argument, 0, 'u'},
			{"testing_only",                  no_argument, 0, 'e'},
			{0, 0, 0, 0}};
		const int opt = getopt_long(argc,argv, "hm:t:z:p:i:r:b:s:c:n:d:a:w:g:u:e",
									long_options,(int *)0);


		if(opt == -1)
			break;

		switch (opt)
		{
		case 'h':
			usage();
			break;
		case 'm':
			arg_metric_type = atoi(optarg);
			break;
		case 't':
			arg_transform_type = atoi(optarg);
			break;
		case 'z':
			arg_optimizer_type = atoi(optarg);
			break;
		case 'p':
			arg_parameters_file = optarg;
			break;
		case 'i':
			arg_max_iter = atoi(optarg);
			break;
		case 'r':
			arg_resolution_levels = atoi(optarg);
			break;
		case 'b':
			arg_histogram_bins = atoi(optarg);
			break;
		case 's':
			arg_grid_spacing = atof(optarg);
			break;
		case 'c':
            arg_control_points_file = optarg;
            break;
		case 'n':
            arg_constrained_points = 1;
            arg_constrained_points_file = optarg;
            break;
		case 'a':
            arg_affine_transform = 1;
            arg_affine_transform_file = optarg;
            break;
		case 'u':
			arg_tau = atof(optarg);
			break;
		case 'e':
			arg_testing_only = 1;
			break;
		case '?':
		default:
			std::cout << "Parse error. Aborting." << std::endl;
			abort();
		}
	}

	if(argc - optind != 2)
	{
		std::cout << "Wrong number of input arguments.\n\n";
		usage();
	}



	char * fixedImageFile               = argv[optind];
	char * movingImageFile              = argv[optind+1];

   
   
	typedef double ScalarType;
  
	const    unsigned int    Dimension = 3;

	typedef  float  PixelType;
  
	typedef itk::Image< PixelType, Dimension >  FixedImageType;
	typedef itk::Image< PixelType, Dimension >  MovingImageType;

	typedef   float     InternalPixelType;

	typedef itk::Image< InternalPixelType, Dimension > InternalImageType;


	typedef itk::ImageFileReader< FixedImageType  > FixedImageReaderType;
	typedef itk::ImageFileReader< MovingImageType > MovingImageReaderType;

	typedef itk::CastImageFilter< 
		FixedImageType, InternalImageType > FixedCastFilterType;
  
	typedef itk::CastImageFilter< 
		MovingImageType, InternalImageType > MovingCastFilterType;




	FixedImageReaderType::Pointer  fixedImageReader  = 
		FixedImageReaderType::New();
	MovingImageReaderType::Pointer movingImageReader = 
		MovingImageReaderType::New();

	fixedImageReader->SetFileName( fixedImageFile );
	movingImageReader->SetFileName( movingImageFile );

	FixedCastFilterType::Pointer fixedCaster   = FixedCastFilterType::New();
	MovingCastFilterType::Pointer movingCaster = MovingCastFilterType::New();

	fixedCaster->SetInput(  fixedImageReader->GetOutput() );
	movingCaster->SetInput( movingImageReader->GetOutput() );

	try
	{
		fixedCaster->Update();
		movingCaster->Update();
	}
	catch( itk::ExceptionObject & excep )
	{
		std::cerr << "ExceptionObject caught" << std::endl;
		std::cerr << excep << std::endl;
		return 1;
	}


	InternalImageType::ConstPointer fixedImage  = fixedCaster->GetOutput();
	InternalImageType::ConstPointer movingImage = movingCaster->GetOutput();






	typedef itk::ImageToImageMetric< 
		InternalImageType, 
		InternalImageType >    MetricType;

	typedef itk::Transform<ScalarType,Dimension,Dimension>   TransformType;

	typedef itk::SingleValuedNonLinearOptimizer       OptimizerType;

	typedef itk::LinearInterpolateImageFunction< 
		InternalImageType,
		ScalarType             > InterpolatorType;


	typedef itk::MultiResolutionImageRegistrationMethod< 
		InternalImageType, 
		InternalImageType    > RegistrationType;



	RegistrationType::Pointer   registration  = RegistrationType::New();
	InterpolatorType::Pointer   interpolator  = InterpolatorType::New();
	OptimizerType::Pointer      optimizer; // will be initialized later
	TransformType::Pointer      transform; // will be initialized later
	MetricType::Pointer         metric; // will be initialized later


//   std::cout.setf(std::ios_base::scientific,std::ios_base::floatfield);
	std::cout.precision(16);


/////////////////////////////////////////////////////
// Setup the transform
/////////////////////////////////////////////////////


	typedef OptimizerType::ScalesType    OptimizerScalesType;
	OptimizerScalesType optimizerScales;

	switch(arg_transform_type)
	{
	case 1:   //  ADRevolveEuler3DTransform transform
	{

		typedef itkek::ADRevolveEuler3DTransform< ScalarType > TransformType2;

		TransformType2::Pointer      transform2     = TransformType2::New();


		// If testing, have the registration itself do the work rather than
		// the initializer.
		if(!arg_testing_only)
		{

			typedef itk::CenteredTransformInitializer<
				TransformType2,
				InternalImageType,
				InternalImageType >  TransformInitializerType;
			TransformInitializerType::Pointer initializer = 
				TransformInitializerType::New();
	   
			initializer->SetTransform(   transform2 );
			initializer->SetFixedImage(  fixedImage );
			initializer->SetMovingImage( movingImage );
			initializer->MomentsOn();
			initializer->InitializeTransform();

		}

		//transform2->SetComputeZYX(true);

		optimizerScales.SetSize( 6 );
		//optimizerScales.SetSize( Dimension*(Dimension+1) );
	   
		ScalarType translationScale = 1.0 / 1000.0;
		//ScalarType translationScale = 1.0;

		int c = 0;
		for(unsigned int i=0;i<3;i++)
			optimizerScales[c++]  = 1.0;

		for(unsigned int i=0;i<3;i++)
			optimizerScales[c++]  = translationScale;
	   

		transform = transform2;

		PRINT("Transform: itkek::ADEuler3DTransform");
		//std::cout << "Transform: itkek::ADEuler3DTransform" << std::endl;
      
	}
	break;  

	case 2:   //  ADRevolveMatrixOffsetTransformBase transform
	{

		typedef itkek::ADRevolveMatrixOffsetTransformBase< ScalarType, 
			Dimension, Dimension> TransformType2;

		TransformType2::Pointer      transform2     = TransformType2::New();



		typedef itk::CenteredTransformInitializer<
			TransformType2,
			InternalImageType,
			InternalImageType >  TransformInitializerType;
		TransformInitializerType::Pointer initializer = 
			TransformInitializerType::New();
	   
		initializer->SetTransform(   transform2 );
		initializer->SetFixedImage(  fixedImage );
		initializer->SetMovingImage( movingImage );
		initializer->MomentsOn();
		initializer->InitializeTransform();


		//optimizerScales.SetSize( 12 );
		optimizerScales.SetSize( Dimension*(Dimension+1) );
	   
		ScalarType translationScale = 1.0 / 1000.0;
		//ScalarType translationScale = 1.0;

		int c = 0;
		for(unsigned int i=0;i<Dimension*Dimension;i++)
			optimizerScales[c++]  = 1.0;

		for(unsigned int i=0;i<Dimension;i++)
			optimizerScales[c++]  = translationScale;
	   

		transform = transform2;

		PRINT("Transform: itkek::ADRevolveMatrixOffsetTransformBase");
		//std::cout << "Transform: itkek::ADRevolveMatrixOffsetTransformBase" << std::endl;
      
	}
	break;  

	case 3: // BSpline
	{

		//  Here we define the parameters of the BSplineDeformableTransform
		//  grid.  We arbitrarily decide to use a grid with $5 \times 5$
		//  nodes within the image.  The reader should note that the BSpline
		//  computation requires a finite support region ( 1 grid node at the
		//  lower borders and 2 grid nodes at upper borders). Therefore in
		//  this example, we set the grid size to be $8 \times 8$ and place
		//  the grid origin such that grid node (1,1) coinicides with the
		//  first pixel in the fixed image.
	 
		const unsigned int SplineOrder = 3;
		typedef itkek::ADRevolveBSplineDeformableTransform<
			double,
			Dimension,
			SplineOrder >     TransformType2;

		TransformType2::Pointer      transform2     = TransformType2::New();


		// Load an affine transform if one was given on the command line
		// Otherwise let the affine transform remain the identity.

		typedef itk::AffineTransform< double, Dimension > 
			AffineTransformType;

		AffineTransformType::Pointer affineTransform;
   
		if(arg_affine_transform)
		{
      
			affineTransform = AffineTransformType::New();

			int ret = itkek::LoadSaveAffineTransform::LoadAffineTransform(
				arg_affine_transform_file,
				affineTransform,0);
			if(ret != 0)
			{
				std::cout << "Warning: could not load affine transform. "
						  << "Using identity instead." << std::endl;
				affineTransform = 0;
			}

		}

		//   affineTransform.Print(std::cout);


		// Get information about the image
		typedef TransformType2::SpacingType SpacingType;
		SpacingType spacing = fixedImage->GetSpacing();

		typedef TransformType2::OriginType OriginType;
		OriginType origin = fixedImage->GetOrigin();;

		FixedImageType::RegionType fixedRegion =fixedImage->GetBufferedRegion();
		FixedImageType::SizeType fixedImageSize = fixedRegion.GetSize();

		// Since we're given the desired grid spacing, and the grid origin is
		// computed from this spacing, we only need to compute the grid size

		typedef TransformType2::RegionType RegionType;
		RegionType bsplineRegion;
		RegionType::SizeType   gridSizeOnImage;
		RegionType::SizeType   gridBorderSize;
		RegionType::SizeType   totalGridSize;

		if(arg_grid_spacing <= 0.0)
		{
			std::cout << "Error: Spacing must be positive" << std::endl;
			abort();
		}

		SpacingType gridSpacing;
		gridSpacing.Fill(arg_grid_spacing);

		for(unsigned int r=0; r<Dimension; r++)
		{
			gridSizeOnImage[r] = static_cast<int>( vnl_math_rnd(
													   static_cast<double>(fixedImageSize[r]-1) * spacing[r]  / 
													   arg_grid_spacing ) );

			origin[r]  -= arg_grid_spacing; 
		}

		gridBorderSize.Fill( 3 );// Border for spline order = 3 ( 1 lower, 2 upper )
		totalGridSize = gridSizeOnImage + gridBorderSize;

		bsplineRegion.SetSize( totalGridSize );


		transform2->SetGridSpacing( gridSpacing );
		transform2->SetGridOrigin( origin );
		transform2->SetGridRegion( bsplineRegion );
  
		if(affineTransform.GetPointer())
			transform2->SetBulkTransform(affineTransform);

		typedef TransformType2::ParametersType     ParametersType;

		const unsigned int numberOfParameters =
			transform2->GetNumberOfParameters();

		// The bspline does not store it parameters locally so we need a
		// container which won't go out of scope after the current block
		// we're in. Now the base class of the bspline has an object called
		// m_Parameters which is not used by the bspline class, but there is
		// no function available for accessing it externally (See the file
		// itkTransform.h. Also the GetParameters() function in this file
		// does not return it either). I therefore created a special function
		// called GetParametersFromBase() to get a reference to this object
		// so we could use it to store the parameters. (The other option was
		// to declare the object on the heap, but this creates a potential
		// memory leak.)

		ParametersType& parameters = transform2->GetParametersFromBase();

		parameters.SetSize( numberOfParameters );

		PRINT("numberOfParameters: " << numberOfParameters);
		PRINT("number Of Control Points: " << numberOfParameters/3);

	 
		parameters.Fill( 0.0 );

		transform2->SetParameters( parameters );

		optimizerScales.SetSize( numberOfParameters );
		optimizerScales.Fill(1.0);
  
	 
		transform = transform2;

		PRINT("Transform: itkek::ADRevolveBSplineDeformableTransform");
		//std::cout << "Transform: itkek::ADBSplineDeformableTransform" << std::endl;

	}
	break;

	case 4: // Thin Plate Spline
	{

		typedef itkek::ADRevolveRBFTransform  TransformType2;
		TransformType2::Pointer  transform2         = TransformType2::New();


		// Load an affine transform if one was given on the command line
		// Otherwise let the affine transform remain the identity.

		typedef itk::AffineTransform< double, Dimension > 
			AffineTransformType;
		AffineTransformType::Pointer affineTransform =
			AffineTransformType::New();

   
		if(arg_affine_transform)
		{
      
			int ret = itkek::LoadSaveAffineTransform::
				LoadAffineTransform(arg_affine_transform_file,affineTransform,0);
			if(ret != 0)
			{
				std::cout << "Warning: could not load affine transform. "
						  << "Using identity instead." << std::endl;
			}

		}

		//affineTransform.Print(std::cout);



		//--------------------------------------------------------------

		// Load the unconstrained control points from the file passed on the
		// command line
		typedef itk::DefaultStaticMeshTraits<double,
			3,
			3,
			double,
			double> PointSetTraitsType;
		typedef itk::PointSet<double, 3, PointSetTraitsType> PointSetType;
		typedef PointSetType::PointType PointType;

		PointSetType::Pointer sourceLandmarks = PointSetType::New();

		if(!arg_control_points_file)
		{
			std::cout << "Error: No control points file specified." << std::endl;
			abort();
		}


		LoadPointSet(arg_control_points_file,sourceLandmarks);

		//--------------------------------------------------------------

		// Load the hard constrained control points
		
		//typedef itkek::ADObjectiveFunction ObjectiveFunctionType;
		TransformType2::FixedDisplacementsContainer listFixedPoints;
		
		if(arg_hard_constrained_points)
		{
			
			std::ifstream pointsfile(arg_hard_constrained_points_file);
			if(!pointsfile) 
			{
				std::cout << "Error: cannot open file" << std::endl;
				abort();
			}
			
			
			TransformType2::FixedDisplacementsContainer::value_type pair;
			
			while( pointsfile >> pair.first >> 
				   pair.second[0] >> pair.second[1] >> pair.second[2])
			{
				listFixedPoints.push_back( pair );
			}
			
			pointsfile.close();
			
		}   
		
		
		//--------------------------------------------------------------
		
		
		transform2->
			Setup(arg_tau,          //double tau,
				  1,          //int k,
				  40,          //int gmresIterations,
				  1.0e-6,          //double gmresTolerance,
				  *sourceLandmarks->GetPoints(),//PointSetPointer sorceLndmrks,
				  affineTransform, //AffineTransformPointer affineTransform
				  true, // bool checkpointingFlag
				  listFixedPoints //const FixedDisplacementsContainer&
				  //  fixedDisplacements
				);
		
		//--------------------------------------------------------------
	 


		optimizerScales.SetSize( 3*sourceLandmarks->GetNumberOfPoints() );
		optimizerScales.Fill(1.0);



		transform = transform2;

		PRINT("Transform: itkek::ADRevolveRBFTransform");
		//std::cout << "Transform: itkek::ADRevolveRBFTransform" << std::endl;


	}
	break;

	default:
	{
		std::cout << "Unrecognized transform" << std::endl;
		exit(1);
	}
   
   
   
	}




/////////////////////////////////////////////////////
// Setup the optimizer
/////////////////////////////////////////////////////

	switch(arg_optimizer_type)
	{
	case 1: // RegularStepGradientDescentOptimizer
	{

		typedef itk::RegularStepGradientDescentOptimizer       OptimizerType2;
		OptimizerType2::Pointer      optimizer2     = OptimizerType2::New();

		optimizer2->SetMaximumStepLength( 4.00 );
		optimizer2->SetMinimumStepLength( .01 );
		optimizer2->SetNumberOfIterations( arg_max_iter );

		optimizer2->SetScales( optimizerScales );

		optimizer2->MinimizeOn();

		optimizer = optimizer2;

		PRINT("Optimizer: itk::RegularStepGradientDescentOptimizer");
		//std::cout << "Optimizer: itk::RegularStepGradientDescentOptimizer" << std::endl;
	}
	break;

	case 2:    // LBFGSB
	{

		typedef itk::LBFGSBOptimizer       OptimizerType2;   

		OptimizerType2::Pointer      optimizer2     = OptimizerType2::New();

		const unsigned int numberOfParameters = 
			transform->GetNumberOfParameters();

		OptimizerType2::BoundSelectionType boundSelect( numberOfParameters );
		OptimizerType2::BoundValueType upperBound( numberOfParameters );
		OptimizerType2::BoundValueType lowerBound( numberOfParameters );

		boundSelect.Fill( 0 );
		upperBound.Fill( 0.0 );
		lowerBound.Fill( 0.0 );

		optimizer2->SetBoundSelection( boundSelect );
		optimizer2->SetUpperBound( upperBound );
		optimizer2->SetLowerBound( lowerBound );

		optimizer2->SetCostFunctionConvergenceFactor( 1e+1 );
		//optimizer2->SetProjectedGradientTolerance( 1.0 );
		optimizer2->SetMaximumNumberOfIterations( arg_max_iter );
		optimizer2->SetMaximumNumberOfEvaluations( 25*arg_max_iter );
		//optimizer2->SetMaximumNumberOfCorrections( 10 );

		optimizer = optimizer2;

		PRINT("Optimizer: itk::LBFGSBOptimizer");
		//std::cout << "Optimizer: itk::LBFGSBOptimizer" << std::endl;

	}
	break;


	default:
	{
		std::cout << "Unrecognized option" << std::endl;
		exit(1);
	}

	}

	// Create the Command observer and register it with the optimizer.
	CommandIterationUpdate::Pointer observer = CommandIterationUpdate::New();
	optimizer->AddObserver( itk::IterationEvent(), observer );




/////////////////////////////////////////////////////
// Setup the metric
/////////////////////////////////////////////////////
   
	switch(arg_metric_type)
	{

	case 1: // SSD with AD and Revolve
	{

		typedef itkek::ADRevolveMeanSquaresImageToImageMetric MetricType2;


		MetricType2::Pointer       metric2          = MetricType2::New();

		metric = metric2;

		PRINT("Metric: itkek::ADRevolveMeanSquaresImageToImageMetric");
		//std::cout << "Metric: itkek::ADRevolveMeanSquaresImageToImageMetric" << std::endl;
	 
	}
	break;

	case 2: // MI with AD and Revolve
	{

		typedef itkek::ADRevolveMattesMutualInformationImageToImageMetric MetricType2;


		MetricType2::Pointer       metric2          = MetricType2::New();


		metric2->SetNumberOfHistogramBins( arg_histogram_bins );

		//InternalImageType::RegionType::SizeType s = 
		//   fixedImage->GetLargestPossibleRegion().GetSize();
		//const unsigned int numberOfSamples = s[0]*s[1]*s[2] / 10;
		//metric2->SetNumberOfSpatialSamples( numberOfSamples );

		//vnl_sample_reseed( 76926294 );
		//vnl_sample_reseed( 769263 );

		metric2->UseAllPixelsOn();

		metric = metric2;

		PRINT("Metric: itkek::ADRevolveMattesMutualInformationImageToImageMetric");
	 
	}
	break;

	case 3: // SSD with AD no revolve
	{
		typedef itkek::ADMeanSquaresImageToImageMetric< 
			InternalImageType, 
			InternalImageType >    MetricType1;

		MetricType1::Pointer        metric1       = MetricType1::New();
		metric = metric1;

		PRINT("Metric: itkek::ADMeanSquaresImageToImageMetric");
	}
	break;  

	case 4: // MI with AD no revolve
	{
		typedef itkek::ADMattesMutualInformationImageToImageMetric<
			InternalImageType, 
			InternalImageType >    MetricType1;
		MetricType1::Pointer        metric1       = MetricType1::New();

		metric1->SetNumberOfHistogramBins( arg_histogram_bins );

		//InternalImageType::RegionType::SizeType s = 
		//   fixedImage->GetLargestPossibleRegion().GetSize();
		//const unsigned int numberOfSamples = s[0]*s[1]*s[2] / 10;
		//metric1->SetNumberOfSpatialSamples( numberOfSamples );

		//vnl_sample_reseed( 76926294 );
		//vnl_sample_reseed( 769263 );

		metric1->UseAllPixelsOn();



		metric = metric1;

		std::cout << "Metric: itkek::ADMattesMutualInformationImageToImageMetric.\n"
				  << "Error. do not use this metric"
				  << std::endl;
		abort();
      
	}
	break;  

	case 5: // SSD without AD or revolve
	{
		typedef itkek::MeanSquaresImageToImageMetric<
			InternalImageType, 
			InternalImageType >    MetricType1;

		MetricType1::Pointer        metric1       = MetricType1::New();
		metric = metric1;

		PRINT("Metric: itkek::MeanSquaresImageToImageMetric");
	}
	break;  

	case 6: // MI without AD or revolve
	{
		typedef itkek::MattesMutualInformationImageToImageMetric<
			InternalImageType, 
			InternalImageType >    MetricType1;
		MetricType1::Pointer        metric1       = MetricType1::New();

		metric1->SetNumberOfHistogramBins( arg_histogram_bins );

		//InternalImageType::RegionType::SizeType s = 
		//   fixedImage->GetLargestPossibleRegion().GetSize();
		//const unsigned int numberOfSamples = s[0]*s[1]*s[2] / 10;
		//metric1->SetNumberOfSpatialSamples( numberOfSamples );

		//vnl_sample_reseed( 76926294 );
		//vnl_sample_reseed( 769263 );

		metric1->UseAllPixelsOn();


		metric = metric1;

		PRINT("Metric: itkek::MattesMutualInformationImageToImageMetric");
	}
	break;  

	default:
	{
		std::cout << "Unrecognized option" << std::endl;
		exit(1);
	}
	}


   


/////////////////////////////////////////////////////
// Setup the registration object
/////////////////////////////////////////////////////



	registration->SetTransform( transform );
	registration->SetMetric( metric );
	registration->SetFixedImage( fixedImage );
	registration->SetMovingImage( movingImage );
	registration->SetFixedImageRegion( fixedImage->GetLargestPossibleRegion() );
	registration->SetOptimizer( optimizer );
	registration->SetInterpolator(  interpolator  );
	registration->SetNumberOfLevels( arg_resolution_levels );
	registration->SetInitialTransformParameters( 
		registration->GetTransform()->GetParameters() );




	// Create the Command interface observer and register it with the
	// registration object.
	typedef RegistrationInterfaceCommand<RegistrationType> CommandType;
	CommandType::Pointer command = CommandType::New();
	registration->AddObserver( itk::IterationEvent(), command );




/////////////////////////////////////////////////////
// Finally run the registration
/////////////////////////////////////////////////////

	PRINT("IJ NumberOfIterations " << arg_max_iter);
	PRINT("IJ NumberOfIndependentVariable " 
		  << registration->GetInitialTransformParameters().size());

	PRINT("\nStarting registration...\n\n");

	time_t start_time=0, end_time=0;
	double elapsed_time=0.0;

	try 
	{ 
		start_time = time(0);

		registration->StartRegistration(); 

		end_time = time(0);
		elapsed_time = difftime( end_time, start_time );
	} 
	catch( itk::ExceptionObject & err ) 
	{ 
		std::cout << "ExceptionObject caught !" << std::endl; 
		std::cout << err << std::endl; 
		//std::cout << registration->GetTransform()->GetParameters() << std::endl;

		return 1;
	} 


	typedef RegistrationType::ParametersType ParametersType;
	ParametersType finalParameters  =   optimizer->GetCurrentPosition();
	//unsigned int numberOfIterations =   optimizer->GetCurrentIteration();
	//ScalarType bestValue            =   optimizer->GetValue();

	std::cout.setf(std::ios_base::scientific,std::ios_base::floatfield);
	std::cout.precision(16);

	//
	// Print out results
	//
	PRINT("Final result: ");
	//PRINT(" Total iterations   = " << numberOfIterations);
	//PRINT(" Metric value       = " << bestValue);
	PRINT(" Position           = " << finalParameters);
	PRINT(" Run time           =  " << elapsed_time       << "  sec  =  "
		  << (elapsed_time/60.0) << "  min  =  "
		  << (elapsed_time/3600.0) << "  hr");
	/*std::cout << "Final result: " << std::endl;
	//std::cout << " Total iterations   = " << numberOfIterations << std::endl;
	//std::cout << " Metric value       = " << bestValue          << std::endl;
	std::cout << " Position           = " << finalParameters  << std::endl;
	std::cout << " Run time           =  " << elapsed_time       << "  sec  =  ";
	std::cout << (elapsed_time/60.0) << "  min  =  ";
	std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;*/




	// Save the transformation parameters to a file

	// First we need to downcast to the proper transform
	{
		typedef itkek::ADRevolveMatrixOffsetTransformBase< ScalarType, 
			Dimension, Dimension> ADTransformType;

		ADTransformType* adTransform = 
			dynamic_cast< ADTransformType* >( transform.GetPointer() );
   
		if(adTransform)
		{
			// Create a smart pointer since the function SaveAffineTransform only
			// takes smart pointers
			ADTransformType::Pointer sptr = adTransform;

			itkek::LoadSaveAffineTransform::SaveAffineTransform(arg_parameters_file, sptr);
			return 0;
		}
	}



	// We're not one of the above transforms. Perhaps we're a bspline transform
	{
		const unsigned int SplineOrder = 3;
		typedef itkek::ADRevolveBSplineDeformableTransform<
			double,
			Dimension,
			SplineOrder >     ADTransformType;


		ADTransformType* adTransform = 
			dynamic_cast< ADTransformType* >( transform.GetPointer() );
   
		if(adTransform)
		{
			adTransform->SaveTransform(arg_parameters_file);
			return 0;
		}
	}


	// We're not one of the above transforms. Perhaps we're a tps transform
	{
		typedef itkek::ADRevolveRBFTransform  ADTransformType;


		ADTransformType* adTransform = 
			dynamic_cast< ADTransformType* >( transform.GetPointer() );
   
		if(adTransform)
			adTransform->SaveTransform(arg_parameters_file);
		else
			return 1;



		// In addition save the control points and their transformed
		// points to a file. Each line has 6 numbers. The first 3 are the
		// coordinates of the source point and the last 3 are those of the
		// transformed point
   
		std::ofstream paramfile("source_and_transformed_points.txt");
		if(!paramfile) 
		{
			std::cout << "Error: could not open file for output" << std::endl;
			abort();
		}

		paramfile.setf(std::ios_base::scientific,std::ios_base::floatfield);
		paramfile.precision(16);

		typedef TransformType::InputPointType PointType;

		const ADTransformType::PointsContainer& sourceLandmarks =
			adTransform->GetSourceLandmarks();

		for(unsigned int i=0;i<sourceLandmarks.size();i++)
		{
			const PointType& in = sourceLandmarks[i];

			const PointType& out = adTransform->TransformPoint(in);
      
			paramfile << in[0] << " " << in[1] << " " << in[2] << " "
					  << out[0] << " " << out[1] << " " << out[2] << "\n";
		}


		return 0;
	}


	// If we reach here then we could not cast to the right transform so issue
	// an error.
	std::cout << "Error: Could not downcast to proper transform. "
		"Unable to save transform. " << std::endl;

	return 1;

}
