#include "itkekADRevolveRBFTransform.h"
#include "itkekMacro.h"

#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkImageRegionIterator.h"
#include "itkImageLinearIteratorWithIndex.h"
#include "itkGradientRecursiveGaussianImageFilter.h"
#include "itkImageRandomConstIteratorWithIndex.h"
#include "itkBSplineInterpolateImageFunction.h"
#include "itkCovariantVector.h"
#include "itkImageRandomConstIteratorWithIndex.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageRegionIterator.h"
#include "itkImageIterator.h"
#include "itkBSplineDeformableTransform.h"
#include "vnl/vnl_math.h"
#include "vnl/vnl_sample.h"
#include <fstream>



extern int trace_flag;


namespace itkek
{


// Define static variables

   int ADRevolveRBFTransform::s_MatrixMultiplicationBlockSize;
   ADRevolveRBFTransform::FixedDisplacementsContainer ADRevolveRBFTransform::s_FixedDisplacements;
   double ADRevolveRBFTransform::s_Tolerance=1.0e-6;
   ADRevolveRBFTransform::RBFMatrixType ADRevolveRBFTransform::s_M;
   int ADRevolveRBFTransform::s_NumberOfPoints=0;
   int ADRevolveRBFTransform::s_MaxGMRESRestartIterations=0;
   int ADRevolveRBFTransform::s_MaxGMRESIterations=0;
   int ADRevolveRBFTransform::s_K=1;
   double ADRevolveRBFTransform::s_Tau=0.0;
   double ADRevolveRBFTransform::s_Tau2=0.0;
   ADRevolveRBFTransform::PointsContainer ADRevolveRBFTransform::s_SourceLandmarks;
   ADRevolveRBFTransform::AffineTransformPointer ADRevolveRBFTransform::s_AffineTransform;
   ADRevolveRBFTransform::AffineMatrixType ADRevolveRBFTransform::s_AffineMatrix;
   ADRevolveRBFTransform::ADMatrixType* ADRevolveRBFTransform::s_X = 0;
   const adouble* ADRevolveRBFTransform::s_ParametersAD = 0;


/////////////////////////////////////////////////////////
// Setup functions
/////////////////////////////////////////////////////////   




adolc::ObjectiveFunctionBase::LabelsContainerType 
ADRevolveRBFTransform::GetLabels() const
{

   // Setup the labels vector
   adolc::ObjectiveFunctionBase::LabelsContainerType labels;

   labels.push_back(Label(0,0,0));

   for(int d=0; d<3; d++)
   {

      labels.push_back(Label(0,1,0));

      {
		  for(int i=0;i<s_NumberOfPoints;i++)
		  {
			  if(i%s_MatrixMultiplicationBlockSize==0)
				  labels.push_back(Label(0,2,0));
		  }
      }

      labels.push_back(Label(0,3,0));

      int j = 1;

      while (j <= s_MaxGMRESIterations) 
      {

		  labels.push_back(Label(0,4,0));

	 for (int i = 0; 
	      i < s_MaxGMRESRestartIterations && j <= s_MaxGMRESIterations; 
	      i++, j++) 
	 {

	    labels.push_back(Label(0,5,0));

	    {
	       for(int ii=0;ii<s_NumberOfPoints;ii++)
	       {
			   if(ii%s_MatrixMultiplicationBlockSize==0)
				   labels.push_back(Label(0,6,0));
	       }
	    }

	    labels.push_back(Label(0,7,0));
	    //labels.push_back(Label(0,8,0));

	 } // end for

	 //labels.push_back(Label(0,9,0));
	 labels.push_back(Label(0,8,0));

	 {
	    for(int ii=0;ii<s_NumberOfPoints;ii++)
	    {
	       if(ii%s_MatrixMultiplicationBlockSize==0)
			   labels.push_back(Label(0,9,0));
	    }
	 }
	    
	 labels.push_back(Label(0,10,0));
	 //labels.push_back(Label(0,13,0));
	 //labels.push_back(Label(0,14,0));
    
      } // end while
	 
   } // end dimension




//    if(0)
//    {
//
//       for(unsigned int i=0;i<labels.size();i++)
// 	 std::cout << i << ":  " 
// 		   << labels[i].Get(0) << "  " 
// 		   << labels[i].Get(1) << "  " 
// 		   << labels[i].Get(2) << std::endl;
//    }

   //PRINT(labels.size());


   return labels;

}



void
ADRevolveRBFTransform::
CopyData(const Superclass* rhs)
{

   if(this == rhs)
      return;


   if(!rhs)
   {
      itkExceptionMacro(<< "Error: rhs is null.");
   }

   ConstPointer rhs_cast =
      dynamic_cast< const Self* >(rhs);

   if(rhs_cast.IsNull())
   {
      itkExceptionMacro(<< "Error: Could not cast to "
	 "ADRevolveTransformBase.");
   }



   int numberInts = sizeof(m_Ints) / sizeof(int);



   for(int i=0;i<numberInts;i++)
      m_Ints[i] = rhs_cast->m_Ints[i];
   
}


void
ADRevolveRBFTransform::
Setup(double tau,
      int k,
      int gmresIterations,
      double gmresTolerance,
      PointsContainer& sourceLandmarks,
      AffineTransformPointer affineTransform,
      bool checkpointingFlag,
      const FixedDisplacementsContainer& fixedDisplacements)
{
   PRINT("\nStarting Setup...\n\n");


   s_Tau = tau;
   s_K = k;
   s_MaxGMRESIterations = gmresIterations;
   s_Tolerance = gmresTolerance;
   s_SourceLandmarks = sourceLandmarks; // Note: a full copy
   s_AffineTransform = affineTransform;
   s_CheckpointingFlag = checkpointingFlag;
   s_FixedDisplacements = fixedDisplacements;

   s_Tau2 = s_Tau*s_Tau;

   if(s_MaxGMRESRestartIterations != 0 && 
      s_MaxGMRESRestartIterations != gmresIterations)
   {
      itkExceptionMacro(<< "Error.  The variable s_MaxGMRESRestartIterations "
	 "can only be set once and cannot be changed again throughout the entire program."
	 "This is because this class uses local static variables.");
   }

   s_MaxGMRESRestartIterations = s_MaxGMRESIterations;

   if(s_NumberOfPoints != 0 && 
      s_NumberOfPoints != static_cast<int>(s_SourceLandmarks.size()))
   {
      itkExceptionMacro(<< "Error.  The variable s_NumberOfPoints "
	 "can only be set once and cannot be changed again throughout the entire program."
	 "This is because this class uses local static variables.");
   }

   s_NumberOfPoints = s_SourceLandmarks.size();

   
   if(s_NumberOfPoints == 0)
   {
      itkExceptionMacro(<< "Error: Number of points is zero.");
   }
   if(s_MaxGMRESRestartIterations == 0)
   {
      itkExceptionMacro(<< "Error: s_MaxGMRESRestartIterations is zero.");
   }



   if(s_K % 2 == 0)
   {
      itkExceptionMacro(<< "Error. s_K must be odd.");
   }

   for(int i=0;i<3;i++)
   {
      for(int j=0;j<3;j++)
	 s_AffineMatrix[i][j] = s_AffineTransform->GetMatrix()[i][j];
      s_AffineMatrix[i][3] = s_AffineTransform->GetOffset()[i];
   }


   this->SetupInterpolationMatrix();



//------------------------------------------------------

   if(s_X)
      delete s_X;
   s_X = new ADMatrixType(3,s_NumberOfPoints);
   if(!s_X)
   {
      itkExceptionMacro(<< "Error. Could not allocate memory.");
   }


// End of adouble allocation
//------------------------------------------------------




   // This calculates how many rows to multiply in a matrix multiplication
   // in a single timestep. 
//   s_MatrixMultiplicationBlockSize = 300000 / s_NumberOfPoints;
   s_MatrixMultiplicationBlockSize = 500;
//   s_NumberOfVoxelsPerTimeStep = 200;


//------------------------------------------------------

   //print out important data about this class.

   EVAL(s_AffineTransform.Print(std::cout));

}


void
ADRevolveRBFTransform::
SetupInterpolationMatrix() const
{

   // setup the M matrix of the  systems: M * X1 = B1
   //                                     M * X2 = B2
   //                                     M * X3 = B3

   s_M.SetupMatrix(s_SourceLandmarks,s_K,s_Tau);


//    s_Precond.SetupPreconditioner(s_M,s_NumberOfPoints,
// 				 s_SourceLandmarks);

}


void
ADRevolveRBFTransform::
Function(adolc::ObjectiveFunctionBase::Label& label)
{

   this->GMRES(label);

}



// this used for debugging
#define MATRIX_MULTIPLY_DEBUG(VECTOR)   for(k=0;k<s_NumberOfPoints;++k)   \
	 {                                                                \
	    const InputPointType& sourcepoint = s_SourceLandmarks[k];     \
 	    tmp[k] = 0.0;                                                 \
            double tmpbf;                                             \
	    for(int r=0; r<s_NumberOfPoints; ++r)                         \
	    {                                                             \
	       tmpbf = this->BasicFunction(sourcepoint,                   \
					   s_SourceLandmarks[r]);                         \
	       tmp[k] += tmpbf*VECTOR[r];                                 \
	    }                                                             \
	 } 


#define MATRIX_MULTIPLY(LABEL,VECTOR)   for(k=0;k<s_NumberOfPoints;++k)   \
	 {                                                                \
	    if(k%s_MatrixMultiplicationBlockSize==0)                      \
	    {                                                             \
               TIMESTEP(LABEL);                                       \
	    }                                                             \
	    const InputPointType& sourcepoint = s_SourceLandmarks[k];     \
 	    tmp[k] = 0.0;                                                 \
            double tmpbf;                                             \
	    for(int r=0; r<s_NumberOfPoints; ++r)                         \
	    {                                                             \
	       tmpbf = this->BasicFunction(sourcepoint,                   \
					   s_SourceLandmarks[r]);                         \
	       tmp[k] += tmpbf*VECTOR[r];                                 \
	    }                                                             \
	 } 


void ADRevolveRBFTransform::GMRES(adolc::ObjectiveFunctionBase::Label& label)
{


   // Setup the variables 
   int& d = m_Ints[0];
   int& i = m_Ints[1];
   int& j = m_Ints[2];
   int& k = m_Ints[3];



   // Global 
   ADMatrixType& x = *s_X;


   static adouble temp = 0.0;
   static adouble beta = 0.0;
   static adouble normb = 0.0;
   static adouble resid = 0.0;
   static adouble normw = 0.0;
   static adouble dotproduct = 0.0;

   static ADVectorType transformedPoint(3);
   static ADVectorType transformedPoint2(3);



   static ADVectorType r(s_NumberOfPoints);
   static ADVectorType w(s_NumberOfPoints);
   static ADVectorType tmp(s_NumberOfPoints);
   static ADVectorType tmp2(s_NumberOfPoints);

   static ADVectorType cs(s_MaxGMRESRestartIterations+1);
   static ADVectorType sn(s_MaxGMRESRestartIterations+1);
   static ADVectorType s(s_MaxGMRESRestartIterations+1);

   static ADMatrixType v(s_MaxGMRESRestartIterations+1, s_NumberOfPoints);
   static ADMatrixType H(s_MaxGMRESRestartIterations+1, s_MaxGMRESRestartIterations+1);
   static ADMatrixType b(3, s_NumberOfPoints);

   
   if(s_CheckpointingFlag)
   {JUMPTABLE_10_GOTO_MACRO;}

  
   TIMESTEP(0);
   //L0:


   {
   int jj=0;
   for(int kk=0;kk<3;kk++)
      for(int ii=0;ii<s_NumberOfPoints;ii++)
      {
		b[ kk ][ii] = s_ParametersAD[jj++]; 
      }

   // Set the points that are fixed. These displacements are fixed and are
   // just dummy independent variables. Hence the derivatives with respect to
   // these variables are always zero.
   for(unsigned int ii=0;ii<s_FixedDisplacements.size();ii++)
   {
      for(int kk=0;kk<3;kk++)
      {
		b[ kk ][ s_FixedDisplacements[ii].first ] = 
	    s_FixedDisplacements[ii].second[kk]; 
      }
   }
   //PRINT(s_FixedDisplacements.size());
   }
 

   for(d=0; d<3; d++)
   {


//GMRES(const RBFMatrixType& A, int d,
//      const int& m, int& max_iter, double& tol,
//      int startiter, int enditer, TStateSpace& state) const
      

      //tol = m_Tolerance;
      j = 1;

   

      //for(int ii=0;ii<m_NumberOfPoints;ii++)
      // x[d][ii] = 0.1;
      x[d] = 0.1;

   
      
      //this->Norm(b,size,normb);
//      m_Precond.solve(b.get_row(d),tmp,temp3);
//      s_Precond.solve(b[d],tmp,temp);
      tmp = b[d];

      this->Norm(tmp,s_NumberOfPoints,normb);


      TIMESTEP(1);
//      return;
//     L1:

      {
		  MATRIX_MULTIPLY(2,x[d]);
      }

 
      TIMESTEP(3);
//      return;
//     L3:


      tmp *= -1.0;
      tmp += b[d];

      //s_Precond.solve(tmp,r,temp);
      r = tmp;

      this->Norm(r,s_NumberOfPoints,beta);

   

      if (normb.value() == 0.0)
	 normb = 1.0;
      
      //if ((resid = norm(r) / normb) <= tol) {
      resid = beta.value() / normb.value();
      /*
      if (resid.value() <= s_Tolerance) 
      {
	 // For simplicity this if statement has been removed.
	 // Thus the only way the iteration will end is by 
	 // max iterations.

      }
      */

      //while (j <= max_iter[d]) 
      while (j <= s_MaxGMRESIterations) 
      {

	 TIMESTEP(4);
   

	 v[0] = r;
	 v[0] *= (1.0/beta);
      
//	 for(int ii=0;ii<=s_MaxGMRESRestartIterations;ii++)
//	    s[ii] = 0.0;
	 s = 0.0;
	 s[0] = beta;
   
      

	 for (i = 0; 
	      i < s_MaxGMRESRestartIterations && j <= s_MaxGMRESIterations; 
	      i++, j++) 
	 {
   

	    TIMESTEP(5);


	    {
	       MATRIX_MULTIPLY(6,v[i]);
	    }


	    TIMESTEP(7);


	    //s_Precond.solve(tmp,w,temp);
	    w = tmp;

   
	    for (k = 0; k <= i; k++) {

	       this->DotProduct(w,v[k],s_NumberOfPoints,dotproduct);
	       H[k][i] = dotproduct;

	       tmp2 = v[k];
	       tmp2 *= H[k][i];
	       w -= tmp2;
	    }


	    //H(i+1, i) = norm(w);
	    this->Norm(w,s_NumberOfPoints,normw);
	    H[i+1][i] = normw;


	    v[i+1] = w;

	    temp = 1.0/H[i+1][i];
	    v[i+1] *= temp;

	    for (k = 0; k < i; k++)
	    {
	       this->ApplyPlaneRotation(H[k][i], H[k+1][i], cs[k], sn[k]);
	    }      


	    this->GeneratePlaneRotation(H[i][i], H[i+1][i], cs[i], sn[i]);

	    this->ApplyPlaneRotation(H[i][i], H[i+1][i], cs[i], sn[i]);

	    this->ApplyPlaneRotation(s[i], s[i+1], cs[i], sn[i]);

      
	    resid=vnl_math_abs(s[i+1].value())/normb.value();
	    /*
	    if (resid.value()< s_Tolerance) 
	    {
	       // For simplicity this if statement has been removed.
	       // Thus the only way the iteration will end is by 
	       // max iterations.
	    }
	    */

	 } // end for


//	 UpdateGMRES(x.get_row(d),s_MaxGMRESRestartIterations-1,H,s,v);
	 UpdateGMRES(x[d],s_MaxGMRESRestartIterations-1,H,s,v);


	 TIMESTEP(8);

	    //this->FastMatrixVectorMultiplication(s_M , x[d] , tmp);
	    {
	       MATRIX_MULTIPLY(9,x[d]);
	    }

	    TIMESTEP(10);


	    tmp *= -1.0;
	    tmp += b[d];



	    //s_Precond.solve(tmp,r,temp);
	 r = tmp;

   

	 //beta = norm(r);
	 this->Norm(r,s_NumberOfPoints,beta);
	 resid = beta.value() / normb.value();


	 /*
	 if (resid.value() < s_Tolerance) {

	       // For simplicity this if statement has been removed.
	       // Thus the only way the iteration will end is by 
	       // max iterations.

	 }
	 */

      } // end while

	    

      //tol = resid.value();

      //PRINT(j);
      //PRINT(d);
      //PRINT(resid.value());


      // Compute the real residual using a full real matrix multiplication.
      // This is only for debugging.
      //if(0)
      if(trace_flag == 0)
      {
      
	 MATRIX_MULTIPLY_DEBUG(x[d]);

	 double residual=0.0;
	 for(int ii=0;ii<s_NumberOfPoints;ii++)
	 {
	    residual += fabs(tmp[ii].value() - b[d][ii].value());
	 }

	 residual /= s_NumberOfPoints;

	 //PRINT(residual);

      }

   }



   return;

}

#undef MATRIX_MULTIPLY_DEBUG
#undef MATRIX_MULTIPLY





void 
ADRevolveRBFTransform::GeneratePlaneRotation(badouble dx, badouble dy,
					   badouble cs, badouble sn) const
{

// Original templates code:

//   if (dy == 0.0) {
//      cs = 1.0;
//      sn = 0.0;
//   } else if (abs(dy) > abs(dx)) {
//      Real temp = dx / dy;
//      sn = 1.0 / sqrt( 1.0 + temp*temp );
//      cs = temp * sn;
//   } else {
//      Real temp = dy / dx;
//      cs = 1.0 / sqrt( 1.0 + temp*temp );
//      sn = temp * cs;
//   }

   
   static adouble temp = 0.0;
   static adouble temp2 = 0.0;


   if (dy.value() == 0.0) {
      cs = 1.0;
      sn = 0.0;
   } else if (vnl_math_abs(dy.value()) > 
	      vnl_math_abs(dx.value())) {
      temp = dx / dy;
      temp2 = temp;
      temp2 *= temp2;
      temp2 += 1.0;
      temp2 = sqrt(temp2);
      sn = 1.0 / temp2;
      cs = temp;
      cs *= sn;
   } else {
      temp = dy / dx;
      temp2 = temp;
      temp2 *= temp2;
      temp2 += 1.0;
      temp2 = sqrt(temp2);
      cs = 1.0 / temp2;
      sn = temp;
      sn *= cs;
   }

}


void 
ADRevolveRBFTransform::ApplyPlaneRotation(badouble dx, badouble dy, 
					badouble cs, badouble sn) const
{

// Original templates code:

   //Real temp  =  cs * dx + sn * dy;
   //dy = -sn * dx + cs * dy;
   //dx = temp;


   static adouble temp = 0.0;
   static adouble temp2 = 0.0;



   temp = cs;
   temp *= dx;
   temp2 = sn;
   temp2 *= dy;
   temp += temp2;

   temp2 = cs;
   temp2 *= dy;
   dy = -1.0;
   dy *= dx;
   dy *= sn;
   dy += temp2;

   dx = temp;
}



void
ADRevolveRBFTransform::UpdateGMRES(ADVectorType& x, const int k, 
				 const ADMatrixType& h, 
				 const ADVectorType& s, 
				 const ADMatrixType& v) const
{
// Original templates code:

//   Vector y(s);
//
//   // Backsolve:
//   for (int i = k; i >= 0; i--) {
//      y(i) /= h(i,i);
//      for (int j = i - 1; j >= 0; j--)
//	 y(j) -= h(j,i) * y(i);
//   }
//
//   for (int j = 0; j <= k; j++)
//      x += v[j] * y(j);




   static ADVectorType y(s_MaxGMRESRestartIterations+1);
   static ADVectorType tmp(s_NumberOfPoints);
      

   y = s;
      

   // Backsolve:  
   for (int i = k; i >= 0; i--) {
      


      y[i] = y[i] / h[i][i];
      //y[i] /= h[i][i];

      for (int j = i - 1; j >= 0; j--)
      {
		  y[j] -= h[j][i] * y[i];
      }

   }
      


   for (int j = 0; j <= k; j++)
   {

      tmp = v[j];
      tmp *= y[j];
      x += tmp;

   }
      
}


void
ADRevolveRBFTransform::
Norm(const ADVectorType& v, const int size, //size no longer used
     adouble& norm) const
{

   norm = v*v;

   norm = sqrt(norm);

}


void
ADRevolveRBFTransform::
DotProduct(const ADVectorType& v1, 
	   const ADVectorType& v2, const int size, //size no longer used
	   adouble& result) const
{

   result = v1 * v2;
   
}







void
ADRevolveRBFTransform::
TransformPoint__(const InputPointType& thisPoint, 
		    ADVectorType& outputPoint) const
{

   static ADVectorType val(s_NumberOfPoints);


   // Add the affine part
   {
   double* tmpPoint = outputPoint.operator_equal_start();
   for(int i=0;i<3;i++)
   {
      tmpPoint[i] = (s_AffineMatrix(i,0)*thisPoint[0]+
			s_AffineMatrix(i,1)*thisPoint[1]+
			s_AffineMatrix(i,2)*thisPoint[2]+
			s_AffineMatrix(i,3));
   }
   outputPoint.operator_equal_end();
   }



   // Do not use the fast multipole method, so instead we just sum all the
   // basis functions
   double* tmp = val.operator_equal_start();
   for(int m=0;m<s_NumberOfPoints;m++)
   {

      tmp[m] = this->BasicFunction(thisPoint, s_SourceLandmarks[m]);

   }
   val.operator_equal_end();


   for(int i=0;i<3;i++)
      outputPoint[i] += val * (*s_X)[i];


}






ADRevolveRBFTransform::OutputPointType 
ADRevolveRBFTransform::
TransformPoint(const InputPointType& point) const
{

   static ADVectorType transformedPoint(3);

   
   this->TransformPoint__(point,transformedPoint);

   
   OutputPointType::ValueType output[3] = 
      {
		  transformedPoint[0].value(),
		  transformedPoint[1].value(),
		  transformedPoint[2].value()
      };



   return OutputPointType(output);
}


void
ADRevolveRBFTransform
::SetParameters( const ParametersType& parameters ) 
{

   s_CheckpointingFlag = false;


   const unsigned int size = parameters.size();

   if(size != static_cast<unsigned>(3*s_NumberOfPoints))
   {
      itkExceptionMacro(<< "Error: number of parameters not equal to 3*s_NumberOfPoints.");
   }

   if(s_NumberOfPoints == 0)
   {
      itkExceptionMacro(<< "Error: Number of points is zero.");
   }


   if(s_ParametersAD)
   {
      adouble* tmp = const_cast<adouble*>(s_ParametersAD);

      for(unsigned int i=0;i<size;i++)
		  tmp[i] = parameters[i];
   }
   else
   {

      if(m_ParametersAD) 
      {
		  itkWarningMacro(<< "Error: All adouble variables must allocated and "
						  "deallocated in a last in first out bases. If you delete "
						  "variables here you will probably be violating the LIFO rule.");
		  delete [] m_ParametersAD;
      }



      m_ParametersAD = new adouble[size];


      for(unsigned int i=0;i<size;i++)
		  m_ParametersAD[i] = parameters[i];

      this->SetParametersAD(m_ParametersAD);
   }

   Label dummy;
   this->GMRES(dummy);



}	 


const ADRevolveRBFTransform::ParametersType& 
ADRevolveRBFTransform
::GetParameters(void) const
{ 

   const unsigned int size = 3*s_NumberOfPoints;

   m_Parameters.SetSize(size);

   if(s_ParametersAD)
   {
      for(unsigned int i=0;i<size;i++)
		  m_Parameters[i] = s_ParametersAD[i].value();
   }
   else
   {
      m_Parameters.Fill(0.0);
   }

   return m_Parameters; 

};


// The transform is saved in the following format:
// where # means a number
/*
NameOfClass
k #
tau #
affine
# # # #
# # # #
# # # #
sourcepoints #
# # # ...
... 
...
displacements
# # # ...
... 
...


*/


void
ADRevolveRBFTransform
::SaveTransform(const char* filename) const
{

   if(!s_ParametersAD)
   {
      itkExceptionMacro(<< "Error: Could not save transform.");
   }

   std::ofstream fout(filename);
   if(!fout)
   {
      itkExceptionMacro(<< "Error: could not open file for saving transform");
   }

   fout.setf(std::ios_base::scientific,std::ios_base::floatfield);
   fout.precision(16);


   fout << this->GetNameOfClass() << "\n";
   fout << "k " << s_K << "\n";
   fout << "tau " << s_Tau << "\n";

   fout << "affine " << "\n";

   for(int i=0; i < 3; i++)
   {
      for(int j=0; j < 3; j++)
      {
	 fout << s_AffineTransform->GetMatrix()[i][j] << "  ";
      }
      fout << s_AffineTransform->GetOffset()[i] << "\n";
   }



   fout << "sourcepoints " << s_NumberOfPoints << "\n";



   for(int i=0;i<s_NumberOfPoints;i++)
   {
      //point = s_SourceLandmarks->GetPoints()->GetElement(i);
      const InputPointType& point = s_SourceLandmarks[i];
      fout << point[0] << " " << point[1] << " " << point[2] << "\n";
   }

   fout << "displacements";

   for(int i = 0;i<3*s_NumberOfPoints;i++)
   {
      if(i%3==0)
		  fout << "\n";
      fout << s_ParametersAD[i].value() << " ";
   }

   fout << "\n";

   fout.close();
}





void
ADRevolveRBFTransform
::LoadTransform(const char* filename)
{

   
   std::ifstream fin(filename);

   if(!fin)
   {
      itkExceptionMacro(<< "Error: could not open file for loading transform");
   }

   std::string words;

   fin >> words;
   if(words != this->GetNameOfClass() )
   {
      itkExceptionMacro(<< "Error reading input file " << words);
   }

   int k;
   fin >> words >> k;

   double tau;
   fin >> words >> tau;



   AffineTransformType::Pointer affineTransform = AffineTransformType::New();


   AffineTransformType::ParametersType transformParameters(12);


   fin >> words;

   for(int i=0;i<12;i++)
   {
      fin >> transformParameters[i];
   }


   AffineTransformType::MatrixType matrix;
   AffineTransformType::OffsetType offset;

   int c=0;
   for(int i=0; i < 3; i++)
   {
      for(int j=0; j < 3; j++)
       {
		   matrix[i][j] = transformParameters[c++];
      }
      offset[i] = transformParameters[c++];
   }

   
   affineTransform->SetMatrix(matrix);
   affineTransform->SetOffset(offset);


   int numberOfPoints;
   fin >> words >> numberOfPoints;

   InputPointType point;


   PointsContainer sourceLandmarks(numberOfPoints);

   for(int i=0;i<numberOfPoints;i++)
   {
      fin >> point[0] >> point[1] >> point[2];

      sourceLandmarks[i] = point;
   }


   ParametersType parameters( 3 * numberOfPoints );


   fin >> words;

   for(int i=0;i<3*numberOfPoints;i++)
   {
      fin >> parameters[i];
   }


   fin.close();



#ifdef  IRAD_VERBOSE_OUTPUT
   std::cout << this->GetNameOfClass() << "\n";
   std::cout << "sourcepoints " << numberOfPoints << "\n";
   std::cout << parameters << std::endl;
   std::cout << std::endl;
#endif

   
   MovingImagePointer movingImage;
   FixedImagePointer fixedImage;
   MovingImageInterpolatorPointer interpolator;



   this->Setup(tau,          //double tau,
			   k,          //int k,
			   40,          //int gmresIterations,
			   1.0e-6,          //double gmresTolerance,
			   sourceLandmarks, //PointSetPointer sourceLandmarks,
			   affineTransform,  //AffineTransformPointer affineTransform
			   false//,            // bool checkpointingFlag
	   );


   this->SetParameters(parameters);




}


}
